"""
Relevant Name Changes:
-   BufferID -> ComponentName
"""

from collections import defaultdict
from typing import Callable, List, Tuple
import islpy as isl

from accelforge.util._frozenset import oset
from accelforge.frontend.mapping import (
    Mapping,
    MappingNode,
    # Iterations
    Loop,
    Spatial,
    Temporal,
    # Splits
    Pipeline,
    Sequential,
    # Logical hardware features
    Storage,
    Compute,
)
from accelforge.frontend.workload import TensorName, Workload

from accelforge.model._looptree.mapping_utilities import get_paths
from accelforge.model._looptree.types import ComponentName
from accelforge.model._looptree.reuse.isl.isl_functions import (
    dim_projector_mask,
    insert_equal_dims_map,
)
from accelforge.model._looptree.reuse.isl.mapping_to_isl import DUMP_ISL_IR

from .types import (
    # Bookkeeping objects
    BufferTensorEinsum,
    ComputeEinsum,
    EinsumName,
    Skew,
    SkewsInfo,
    # Tags
    Tag,
    TemporalTag,
    SpatialTag,
    PipelineTag,
    SequentialTag,
)


def skews_from_mapping(mapping: Mapping, workload: Workload) -> SkewsInfo:
    """
    Given a mapping and workload, compute the skew relationships for buffers and
    computes.
    TODO: Fill this in with more accurate description.

    Parameters
    ----------
    mapping:
        The mapping being analyzed.
    workload:
        The workload being executed.

    Returns
    -------
    Skew information for buffer-tensor-einsum and compute-einsum combinations.
    """
    compute_einsum_to_skew: dict[ComputeEinsum, Skew] = defaultdict()
    buffer_tensor_einsum_to_skew: dict[BufferTensorEinsum, Skew] = defaultdict()

    for path in get_paths(mapping):
        leaf: Compute = path[-1]

        # Get the last storage node in path for a particular buffet.
        buffer_to_last_storage_node: dict[ComponentName, MappingNode] = {}
        buffer_node: List[Tuple[ComponentName, MappingNode]] = []
        all_buffer_tensors: List[Tuple[ComponentName, TensorName]] = []

        node: MappingNode
        for node in path:
            match node:
                case Storage():
                    buffer: ComponentName = node.component
                    buffer_to_last_storage_node[buffer] = node
                    buffer_node.append((buffer, node))
                    # TODO: Check this is correct
                    all_buffer_tensors.extend(
                        (buffer, tensor) for tensor in node.tensors
                    )
                case Compute():
                    compute: ComponentName = node.component
                    buffer_to_last_storage_node[compute] = node
                    buffer_node.append((compute, node))

        node_to_current_buffer: dict[MappingNode, MappingNode] = {}
        buffer_idx: int = 0
        for node in path:
            _, cur_buf_last_node = buffer_node[buffer_idx]
            node_to_current_buffer[node] = cur_buf_last_node

            if node == cur_buf_last_node:
                buffer_idx += 1

        # Generate tags, map, and which dims (and tags) should be removed per buffer.
        tags: List[Tag] = []
        base_space: str = f"{leaf.component}_spacetime"
        removal_map: isl.Map = (
            isl.Map.from_multi_aff(
                isl.MultiAff.identity_on_domain_space(
                    isl.Space.alloc(isl.DEFAULT_CONTEXT, 0, 0, 0).domain()
                )
            )
            .set_tuple_name(isl.dim_type.in_, base_space)
            .set_tuple_name(isl.dim_type.out, base_space)
        )

        buffer_storage_past: set[Tuple[ComponentName, TensorName]] = oset()
        buffer_fully_complete: set[ComponentName] = oset()
        buffer_to_dim_removal_mask: defaultdict[
            Tuple[ComponentName, TensorName], List[bool]
        ] = defaultdict(list)

        def add_tag(
            tag: Tag,
            mask_condition: Callable[[ComponentName, TensorName], bool] = (
                lambda b, t: b in buffer_fully_complete
            ),
        ) -> None:
            """
            Performs necessary modifications to removal_map and removal_mask to
            accommodate tagging.

            Parameters
            ----------
            tag:
                The tag to add.
            mask_condition:
                Boolean resolution for the removal mask.

            Postconditions
            --------------
            -   `tags` has another tag appended to it.
            -   `removal_map` has an input and output dimension added that are equal
                to each other.
            -   `removal_mask` has a new entry.
            """
            nonlocal tags
            tags.append(tag)
            nonlocal removal_map
            removal_map = insert_equal_dims_map(
                removal_map,
                removal_map.dim(isl.dim_type.in_),
                removal_map.dim(isl.dim_type.out),
                1,
            )
            if DUMP_ISL_IR:
                print(f"skew removal_map: {removal_map}")
                print(f"tag: {tag}")

            nonlocal all_buffer_tensors
            nonlocal buffer_to_dim_removal_mask
            for buffer_tensor in all_buffer_tensors:
                removal_mask = buffer_to_dim_removal_mask[buffer_tensor]
                removal_mask.append(mask_condition(*buffer_tensor))

        for node in path:
            match node:
                case Storage():
                    buffer_storage_past.update(
                        (node.component, tensor) for tensor in node.tensors
                    )
                    if node == buffer_to_last_storage_node[node.component]:
                        buffer_fully_complete.add(node.component)
                case Loop():
                    tag: Tag
                    if isinstance(node, Temporal):
                        tag: Tag = TemporalTag()
                    elif isinstance(node, Spatial):
                        tag: Tag = SpatialTag(0, node_to_current_buffer[node])
                    else:
                        raise ValueError(
                            f"Type {type(node)} is an iteration not in space or time."
                        )

                    # TODO: Verify logical equivalence to:
                    # https://github.com/NVlabs/timeloop/blob/32370826fdf1aa3c8deb0c93e6b2a2fc7cf053aa/src/loop-analysis/mapping-to-isl/fused-mapping-to-isl.cpp#L660-L671
                    add_tag(
                        tag,
                        lambda b, t: (
                            (b in buffer_fully_complete)
                            or (
                                (b, t) in buffer_storage_past
                                and isinstance(node, Temporal)
                            )
                        ),
                    )
                case Pipeline():
                    add_tag(PipelineTag())
                case Sequential():
                    add_tag(SequentialTag())

        for buffer_tensor in all_buffer_tensors:
            mask: List[bool] = buffer_to_dim_removal_mask[buffer_tensor]
            domain: isl.Set = removal_map.domain()
            projector: isl.Map = dim_projector_mask(domain.get_space(), mask)
            removal_projection: isl.Map = projector.apply_range(removal_map)
            # Attach tuple names per-buffer so downstream occupancy maps keep the spacetime label.
            space_name: str = f"{buffer_tensor[0]}_spacetime"
            removal_projection = removal_projection.set_tuple_name(
                isl.dim_type.in_, space_name
            ).set_tuple_name(isl.dim_type.out, space_name)

            buffer_tags: List[Tag] = [tag for i, tag in enumerate(tags) if not mask[i]]

            # TODO: This buffet structure makes no sense in this context:
            # https://github.com/NVlabs/timeloop/blob/32370826fdf1aa3c8deb0c93e6b2a2fc7cf053aa/src/loop-analysis/mapping-to-isl/fused-mapping-to-isl.cpp#L740-L743
            buffer_tensor_einsum_to_skew[BufferTensorEinsum(*buffer_tensor, leaf)] = (
                Skew(buffer_tags, removal_projection)
            )

        # TODO: Figure out what is actually:
        # https://github.com/NVlabs/timeloop/blob/32370826fdf1aa3c8deb0c93e6b2a2fc7cf053aa/src/loop-analysis/mapping-to-isl/fused-mapping-to-isl.cpp#L746
        compute_einsum_to_skew[ComputeEinsum(leaf.component, leaf)] = Skew(
            tags, removal_map
        )
        einsum: EinsumName = leaf.einsum
        for tensor in workload.einsums[einsum].input_tensor_names:
            buffer_tensor_einsum_to_skew[
                BufferTensorEinsum(leaf.component, tensor, leaf)
            ] = Skew(tags, removal_map)

        for tensor in workload.einsums[einsum].output_tensor_names:
            buffer_tensor_einsum_to_skew[
                BufferTensorEinsum(leaf.component, tensor, leaf)
            ] = Skew(tags, removal_map)

    return SkewsInfo(buffer_tensor_einsum_to_skew, compute_einsum_to_skew)



"""
Tests for workload parsing functions:
  - _parse_einsum_string
  - _parse_einsum_entry
  - _parse_projection
  - _projection_factory
  - isl_expression_has_variable
"""

import unittest

from accelforge.frontend.workload import (
    _parse_einsum_string,
    _parse_einsum_entry,
    _parse_projection,
    _projection_factory,
    isl_expression_has_variable,
    ImpliedProjection,
    TensorAccess,
)

# ============================================================================
# _parse_projection
# ============================================================================


class TestParseProjection(unittest.TestCase):
    """Tests for _parse_projection which converts a projection string to dict/list."""

    def test_simple_identifiers_become_dict(self):
        result = _parse_projection("a, b, c")
        self.assertEqual(result, {"A": "a", "B": "b", "C": "c"})

    def test_explicit_equals_projection(self):
        result = _parse_projection("B:b, M:p, H:h, E:e")
        self.assertEqual(result, {"B": "b", "M": "p", "H": "h", "E": "e"})

    def test_mixed_implicit_and_explicit(self):
        result = _parse_projection("b, M:p, h, e")
        self.assertEqual(result, {"B": "b", "M": "p", "H": "h", "E": "e"})

    def test_single_element(self):
        result = _parse_projection("x")
        self.assertEqual(result, {"X": "x"})

    def test_single_explicit(self):
        result = _parse_projection("X:y")
        self.assertEqual(result, {"X": "y"})

    def test_empty_raises_value_error(self):
        with self.assertRaises(ValueError):
            _parse_projection("")

    def test_invalid_element_raises(self):
        with self.assertRaises(ValueError):
            _parse_projection("123")

    def test_element_with_operator_raises(self):
        with self.assertRaises(ValueError):
            _parse_projection("a+b")

    def test_trailing_comma_raises(self):
        with self.assertRaises(ValueError):
            _parse_projection("a, b, ")

    def test_double_comma_raises(self):
        with self.assertRaises(ValueError):
            _parse_projection("a,, b")

    def test_spaces_stripped(self):
        result = _parse_projection("  a ,  b ,  c  ")
        self.assertEqual(result, {"A": "a", "B": "b", "C": "c"})


# ============================================================================
# _projection_factory
# ============================================================================


class TestProjectionFactory(unittest.TestCase):
    """Tests for _projection_factory which handles list/dict projection input."""

    def test_list_projection_creates_implied(self):
        result = _projection_factory(["a", "b", "c"])
        self.assertIsInstance(result, ImpliedProjection)
        self.assertEqual(result, {"A": "a", "B": "b", "C": "c"})

    def test_dict_projection_returned_as_is(self):
        d = {"M": "m", "N": "n"}
        result = _projection_factory(d)
        self.assertEqual(result, {"M": "m", "N": "n"})
        self.assertNotIsInstance(result, ImpliedProjection)

    def test_invalid_type_raises(self):
        with self.assertRaises(TypeError):
            _projection_factory(123)

    def test_invalid_type_string_raises(self):
        with self.assertRaises(TypeError):
            _projection_factory("abc")

    def test_list_with_non_string_raises(self):
        with self.assertRaises(TypeError):
            _projection_factory([1, 2, 3])

    def test_list_with_expression_raises(self):
        with self.assertRaises(ValueError):
            _projection_factory(["a+b"])

    def test_dict_with_non_string_key_raises(self):
        with self.assertRaises(TypeError):
            _projection_factory({1: "a"})

    def test_dict_with_non_identifier_key_raises(self):
        with self.assertRaises(ValueError):
            _projection_factory({"1invalid": "a"})

    def test_empty_list(self):
        result = _projection_factory([])
        self.assertIsInstance(result, ImpliedProjection)
        self.assertEqual(len(result), 0)

    def test_empty_dict(self):
        result = _projection_factory({})
        self.assertEqual(len(result), 0)


# ============================================================================
# isl_expression_has_variable
# ============================================================================


class TestIslExpressionHasVariable(unittest.TestCase):
    """Tests for isl_expression_has_variable."""

    def test_simple_variable_present(self):
        self.assertTrue(isl_expression_has_variable("0 <= m < 10", "m"))

    def test_simple_variable_absent(self):
        self.assertFalse(isl_expression_has_variable("0 <= m < 10", "n"))

    def test_variable_in_expression(self):
        self.assertTrue(isl_expression_has_variable("a + b < 10", "a"))
        self.assertTrue(isl_expression_has_variable("a + b < 10", "b"))

    def test_operator_not_matched(self):
        """ISL operators like EQ, NE, etc. should not be matched."""
        self.assertFalse(isl_expression_has_variable("EQ", "EQ"))
        self.assertFalse(isl_expression_has_variable("NE", "NE"))
        self.assertFalse(isl_expression_has_variable("AND", "AND"))

    def test_partial_match_not_confused(self):
        """Variable 'm' should not match 'mm' or 'my_var'."""
        self.assertFalse(isl_expression_has_variable("mm + 5", "m"))

    def test_underscore_variable(self):
        self.assertTrue(isl_expression_has_variable("my_var < 5", "my_var"))

    def test_multiple_occurrences(self):
        self.assertTrue(isl_expression_has_variable("m + m < 10", "m"))


# ============================================================================
# _parse_einsum_string
# ============================================================================


class TestParseEinsumString(unittest.TestCase):
    """Tests for _parse_einsum_string."""

    def test_basic_matmul(self):
        result = _parse_einsum_string("C[m, n] = A[m, k] * B[k, n]")
        self.assertEqual(result["name"], "C")
        self.assertEqual(len(result["tensor_accesses"]), 3)

        inputs = [t for t in result["tensor_accesses"] if not t["output"]]
        outputs = [t for t in result["tensor_accesses"] if t["output"]]
        self.assertEqual(len(inputs), 2)
        self.assertEqual(len(outputs), 1)
        self.assertEqual(outputs[0]["name"], "C")

    def test_three_input_tensors(self):
        result = _parse_einsum_string("Y[b] = A[b] * B[b] * C[b]")
        inputs = [t for t in result["tensor_accesses"] if not t["output"]]
        self.assertEqual(len(inputs), 3)
        self.assertEqual([t["name"] for t in inputs], ["A", "B", "C"])

    def test_single_input_tensor(self):
        result = _parse_einsum_string("O[b, m] = I[b, m]")
        self.assertEqual(result["name"], "O")
        self.assertEqual(len(result["tensor_accesses"]), 2)

    def test_dict_projection(self):
        result = _parse_einsum_string(
            "QK[b, m, p, h] = Q[b, m, h, e] * K[B:b, M:p, H:h, E:e]"
        )
        k_tensor = result["tensor_accesses"][1]
        self.assertEqual(k_tensor["name"], "K")
        self.assertEqual(
            k_tensor["projection"], {"B": "b", "M": "p", "H": "h", "E": "e"}
        )

    def test_mixed_projection(self):
        result = _parse_einsum_string(
            "QK[b, m, p, h] = Q[b, m, h, e] * K[b, M:p, h, e]"
        )
        k_tensor = result["tensor_accesses"][1]
        expected = {"B": "b", "M": "p", "H": "h", "E": "e"}
        self.assertEqual(k_tensor["projection"], expected)

    def test_copy_operation(self):
        result = _parse_einsum_string("I[b, m, d] = I_in[b, m, d]")
        self.assertEqual(result["name"], "I")
        self.assertEqual(result["tensor_accesses"][0]["name"], "I_in")

    def test_underscore_in_name(self):
        result = _parse_einsum_string("QK_softmax[b] = QK[b]")
        self.assertEqual(result["name"], "QK_softmax")

    def test_whitespace_handling(self):
        result = _parse_einsum_string("  C[m, n]  =  A[m, k]  *  B[k, n]  ")
        self.assertEqual(result["name"], "C")
        self.assertEqual(len(result["tensor_accesses"]), 3)

    def test_empty_string_raises(self):
        with self.assertRaises(ValueError) as ctx:
            _parse_einsum_string("")
        self.assertIn("cannot be empty", str(ctx.exception))

    def test_whitespace_only_raises(self):
        with self.assertRaises(ValueError) as ctx:
            _parse_einsum_string("   ")
        self.assertIn("cannot be empty", str(ctx.exception))

    def test_no_equals_raises(self):
        with self.assertRaises(ValueError):
            _parse_einsum_string("V[b, m, h, e]")

    def test_no_brackets_raises(self):
        with self.assertRaises(ValueError):
            _parse_einsum_string("V = I * WV")

    def test_no_input_tensors_raises(self):
        with self.assertRaises(ValueError):
            _parse_einsum_string("V[b] =")

    def test_empty_projection_raises(self):
        with self.assertRaises(ValueError):
            _parse_einsum_string("V[] = I[b]")

    def test_tensor_name_starting_with_number_raises(self):
        with self.assertRaises(ValueError):
            _parse_einsum_string("2V[b] = I[b]")

    def test_curly_braces_raise(self):
        with self.assertRaises(ValueError):
            _parse_einsum_string("V{b, m} = I{b}")

    def test_unmatched_brackets_raise(self):
        with self.assertRaises(ValueError):
            _parse_einsum_string("V[b, m = I[b] * W[m]")

    def test_nested_brackets_raise(self):
        with self.assertRaises(ValueError):
            _parse_einsum_string("V[[b]] = I[b]")

    def test_numbers_in_tensor_name_ok(self):
        result = _parse_einsum_string("T2[b] = T1[b] * W1[b]")
        self.assertEqual(result["name"], "T2")

    def test_very_long_tensor_name(self):
        long = "A" * 100
        result = _parse_einsum_string(f"{long}[b] = I[b]")
        self.assertEqual(result["name"], long)


# ============================================================================
# _parse_einsum_entry
# ============================================================================


class TestParseEinsumEntry(unittest.TestCase):
    """Tests for _parse_einsum_entry."""

    def test_entry_with_einsum_key(self):
        entry = {"einsum": "V[b, m, h, e] = I[b, m, d] * WV[h, e, d]"}
        result = _parse_einsum_entry(entry)
        self.assertEqual(result["name"], "V")
        self.assertEqual(len(result["tensor_accesses"]), 3)

    def test_entry_with_einsum_and_renames(self):
        entry = {
            "einsum": "QK[b, m, p, h] = Q[b, m, h, e] * K[b, M:p, h, e]",
            "renames": {"weight": "K", "input": "Q"},
        }
        result = _parse_einsum_entry(entry)
        self.assertEqual(result["name"], "QK")
        self.assertIn("weight", result["renames"])
        self.assertIn("input", result["renames"])

    def test_entry_with_einsum_and_tensor_accesses_extras(self):
        entry = {
            "einsum": "I[b, m, d] = I_in[b, m, d]",
            "tensor_accesses": [{"name": "I_in", "bits_per_value": 16}],
        }
        result = _parse_einsum_entry(entry)
        self.assertEqual(result["name"], "I")
        i_in = [t for t in result["tensor_accesses"] if t["name"] == "I_in"][0]
        self.assertEqual(i_in["bits_per_value"], 16)

    def test_entry_without_einsum_key_passes_through(self):
        entry = {
            "name": "V",
            "tensor_accesses": [
                {"name": "I", "projection": ["b", "m", "d"]},
                {"name": "V", "projection": ["b", "m"], "output": True},
            ],
        }
        result = _parse_einsum_entry(entry)
        self.assertEqual(result["name"], "V")

    def test_invalid_type_raises(self):
        with self.assertRaises(ValueError):
            _parse_einsum_entry(123)

        with self.assertRaises(ValueError):
            _parse_einsum_entry([1, 2, 3])

        with self.assertRaises(ValueError):
            _parse_einsum_entry(None)

    def test_conflicting_tensor_access_attribute_raises(self):
        """If the einsum string sets a field and the extra tensor_accesses also set it, error."""
        entry = {
            "einsum": "V[b, m] = I[b, m]",
            "tensor_accesses": [{"name": "I", "projection": {"B": "b", "M": "m"}}],
        }
        with self.assertRaises(ValueError) as ctx:
            _parse_einsum_entry(entry)
        self.assertIn("already set", str(ctx.exception))

    def test_tensor_access_not_in_einsum_raises(self):
        entry = {
            "einsum": "V[b] = I[b]",
            "tensor_accesses": [{"name": "NotHere", "bits_per_value": 8}],
        }
        with self.assertRaises(ValueError) as ctx:
            _parse_einsum_entry(entry)
        self.assertIn("not found", str(ctx.exception))

    def test_tensor_access_missing_name_raises(self):
        entry = {
            "einsum": "V[b] = I[b]",
            "tensor_accesses": [{"bits_per_value": 8}],
        }
        with self.assertRaises(ValueError) as ctx:
            _parse_einsum_entry(entry)
        self.assertIn("missing a name", str(ctx.exception))

    def test_invalid_tensor_accesses_type_raises(self):
        entry = {
            "einsum": "V[b] = I[b]",
            "tensor_accesses": "not_a_list",
        }
        with self.assertRaises(ValueError):
            _parse_einsum_entry(entry)


if __name__ == "__main__":
    unittest.main()



"""
End-to-end integration tests for values_per_action / bits_per_action precedence.

Each test loads a Toll-based arch with a different vpa/bpa configuration, runs
the mapper, and checks the resulting total energy against the closed-form
expectation computed from the precedence-selected vpa.
"""

import unittest
from pathlib import Path

from accelforge.frontend.spec import Spec
from accelforge.frontend.mapper.metrics import Metrics

TESTS_DIR = Path(__file__).resolve().parent
INPUT_FILES_DIR = TESTS_DIR / "input_files"
WORKLOAD = INPUT_FILES_DIR / "vpa_precedence.workload.yaml"

# Matmul1: T0[m,n0] + W0[n0,n1] + T1[m,n1] with M=128, N0=64, N1=128.
# Every tensor element traverses the Toll exactly once.
TOTAL_ELEMENTS = 128 * 64 + 64 * 128 + 128 * 128  # 32768
READ_ENERGY = 100


def _energy_for_vpa(vpa):
    return TOTAL_ELEMENTS / vpa * READ_ENERGY


class TestValuesPerActionIntegration(unittest.TestCase):
    def _run(self, arch_yaml):
        spec = Spec.from_yaml(str(arch_yaml), str(WORKLOAD))
        spec.mapper.metrics = Metrics.ENERGY
        mappings = spec.map_workload_to_arch(print_progress=False)
        self.assertGreater(
            len(mappings.data), 0, "mapper should return at least one mapping"
        )
        return mappings.data.iloc[0]["Total<SEP>energy"]

    def test_action_vpa_wins(self):
        """action.values_per_action overrides component.values_per_action."""
        energy = self._run(INPUT_FILES_DIR / "vpa_precedence_action_wins.arch.yaml")
        self.assertAlmostEqual(energy, _energy_for_vpa(2))

    def test_component_vpa_used_when_action_empty(self):
        """component.values_per_action used when action's is empty."""
        energy = self._run(INPUT_FILES_DIR / "vpa_precedence_component_wins.arch.yaml")
        self.assertAlmostEqual(energy, _energy_for_vpa(5))

    def test_action_bpa_wins_when_both_vpa_empty(self):
        """action.bits_per_action overrides component.bits_per_action; vpa = bpa/bpv."""
        energy = self._run(INPUT_FILES_DIR / "vpa_precedence_action_bpa.arch.yaml")
        # action.bpa=8, component.bpv=4 -> vpa=2
        self.assertAlmostEqual(energy, _energy_for_vpa(2))

    def test_component_bpa_used_when_action_bpa_and_vpa_empty(self):
        """component.bits_per_action used when nothing else set; vpa = bpa/bpv."""
        energy = self._run(INPUT_FILES_DIR / "vpa_precedence_component_bpa.arch.yaml")
        # component.bpa=16, component.bpv=4 -> vpa=4
        self.assertAlmostEqual(energy, _energy_for_vpa(4))

    def test_default_bpa_is_one(self):
        """Nothing set: action.bpa defaults to 1; vpa = 1/workload_bpv."""
        energy = self._run(INPUT_FILES_DIR / "vpa_precedence_default.arch.yaml")
        # default bpa=1, workload bpv=8 -> vpa = 0.125
        self.assertAlmostEqual(energy, _energy_for_vpa(0.125))


if __name__ == "__main__":
    unittest.main()



from ._symbolic import (
    PRINT_FORMULAS,
    BuffetStats,
    NetworkStats,
    ComputeStats,
    SymbolicAnalysisOutput,
    label_fused_loops,
    quick_insert_reservation_nodes,
    analyze_reuse_and_add_reservations_to_mapping,
)



import unittest

from accelforge.frontend.spec import Spec
from accelforge.model.main import evaluate_mapping, InvalidMappingError
from accelforge.util.parallel import set_n_parallel_jobs

try:
    from .paths import EXAMPLES_DIR
except ImportError:
    from paths import EXAMPLES_DIR

set_n_parallel_jobs(1)


class TestInvalidMapping(unittest.TestCase):
    def test_matmul_to_simple(self):
        M = 64
        KN = 64
        spec = Spec.from_yaml(
            EXAMPLES_DIR / "arches" / "simple.yaml",
            EXAMPLES_DIR / "workloads" / "matmuls.yaml",
            "tests/input_files/mapping/invalid_matmul_to_simple.yaml",
            jinja_parse_data={"N_EINSUMS": 1, "M": M, "KN": KN},
        )
        self.assertRaises(InvalidMappingError, lambda: evaluate_mapping(spec))


class TestModel(unittest.TestCase):
    def test_one_matmul(self):
        M = 64
        KN = 32
        BITS_PER_VALUE = 8
        spec = Spec.from_yaml(
            EXAMPLES_DIR / "arches" / "simple.yaml",
            EXAMPLES_DIR / "workloads" / "matmuls.yaml",
            EXAMPLES_DIR / "mappings" / "unfused_matmuls_to_simple.yaml",
            jinja_parse_data={"N_EINSUMS": 1, "M": M, "KN": KN},
        )

        result = evaluate_mapping(spec)
        energy_breakdown = result.energy(per_component=True, per_tensor=True)
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "T0")], M * KN * BITS_PER_VALUE
        )
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "T1")], M * KN * BITS_PER_VALUE
        )
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "W0")], KN**2 * BITS_PER_VALUE
        )

    def test_bits_per_value_directly_sets(self):
        """bits_per_value on a memory directly sets the bpv for specific tensors."""
        M = 64
        KN = 32
        WORKLOAD_BPV = 8
        OVERRIDE_BPV = 4
        spec = Spec.from_yaml(
            EXAMPLES_DIR / "arches" / "simple.yaml",
            EXAMPLES_DIR / "workloads" / "matmuls.yaml",
            EXAMPLES_DIR / "mappings" / "unfused_matmuls_to_simple.yaml",
            jinja_parse_data={"N_EINSUMS": 1, "M": M, "KN": KN},
        )

        # Set bits_per_value on MainMemory to override T0's bpv to 4
        spec.arch["MainMemory"].bits_per_value = {"T0": OVERRIDE_BPV}

        result = evaluate_mapping(spec)
        energy_breakdown = result.energy(per_component=True, per_tensor=True)
        # T0 should use OVERRIDE_BPV (4) instead of WORKLOAD_BPV (8)
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "T0")], M * KN * OVERRIDE_BPV
        )
        # T1 and W0 should still use the workload's bits_per_value (8)
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "T1")], M * KN * WORKLOAD_BPV
        )
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "W0")], KN**2 * WORKLOAD_BPV
        )

    def test_bits_per_value_all_tensors(self):
        """bits_per_value with All sets bpv for every tensor."""
        M = 64
        KN = 32
        OVERRIDE_BPV = 16
        spec = Spec.from_yaml(
            EXAMPLES_DIR / "arches" / "simple.yaml",
            EXAMPLES_DIR / "workloads" / "matmuls.yaml",
            EXAMPLES_DIR / "mappings" / "unfused_matmuls_to_simple.yaml",
            jinja_parse_data={"N_EINSUMS": 1, "M": M, "KN": KN},
        )

        spec.arch["MainMemory"].bits_per_value = {"All": OVERRIDE_BPV}

        result = evaluate_mapping(spec)
        energy_breakdown = result.energy(per_component=True, per_tensor=True)
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "T0")], M * KN * OVERRIDE_BPV
        )
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "T1")], M * KN * OVERRIDE_BPV
        )
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "W0")], KN**2 * OVERRIDE_BPV
        )

    def test_skip_initial_output_write_false(self):
        M = 64
        KN = 32
        BITS_PER_VALUE = 8
        spec = Spec.from_yaml(
            EXAMPLES_DIR / "arches" / "simple.yaml",
            EXAMPLES_DIR / "workloads" / "matmuls.yaml",
            EXAMPLES_DIR / "mappings" / "unfused_matmuls_to_simple.yaml",
            jinja_parse_data={"N_EINSUMS": 1, "M": M, "KN": KN},
        )

        spec.arch["MainMemory"].skip_initial_output_write = False

        result = evaluate_mapping(spec)
        energy_breakdown = result.energy(per_component=True, per_tensor=True)
        # Input and weight energy unchanged
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "T0")], M * KN * BITS_PER_VALUE
        )
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "W0")], KN**2 * BITS_PER_VALUE
        )
        # Output energy doubles: the initial fill from MainMemory is no longer
        # skipped, so MainMemory sees both the fill read and the writeback write.
        self.assertAlmostEqual(
            energy_breakdown[("MainMemory", "T1")], 2 * M * KN * BITS_PER_VALUE
        )

    def test_two_matmuls(self):
        spec = Spec.from_yaml(
            EXAMPLES_DIR / "arches" / "simple.yaml",
            EXAMPLES_DIR / "workloads" / "matmuls.yaml",
            EXAMPLES_DIR / "mappings" / "unfused_matmuls_to_simple.yaml",
            jinja_parse_data={"N_EINSUMS": 2, "M": 64, "KN": 64},
        )

        result = evaluate_mapping(spec)


if __name__ == "__main__":
    unittest.main()



import functools
from math import ceil, comb
import functools
import math
import numbers
import symengine as se
from accelforge._accelerated_imports import pandas as pd
from accelforge._accelerated_imports import numpy as np
from accelforge.util._frozenset import oset


def _is_numeric(x) -> bool:
    # symengine scalars (RealDouble, Integer, ...) don't register as
    # numbers.Real even though they have a working __float__.
    return isinstance(x, (numbers.Real, se.Basic))


def _is_intlike(x) -> bool:
    if isinstance(x, numbers.Integral):
        return True
    if isinstance(x, (numbers.Real, se.Basic)):
        f = float(x)
        return math.isnan(f) or int(f) == f
    return False


NUMPY_FLOAT_TYPE = np.float32


def _prime_factorization(n):
    f = []
    i = 2
    while n > 1:
        if n % i == 0:
            f.append(i)
            n //= i
        else:
            i += 1
    return f


@functools.lru_cache(maxsize=None)
def _divisors(n):
    return tuple(d for d in range(1, n + 1) if n % d == 0)


@functools.lru_cache(maxsize=None)
def _count_factorizations(n, imperfect_per_loop: tuple[bool, ...]):
    if len(imperfect_per_loop) <= 1:
        return 1

    others = imperfect_per_loop[1:]
    if imperfect_per_loop[0]:
        return sum(_count_factorizations(ceil(n / s), others) for s in range(1, n + 1))
    return sum(_count_factorizations(n // d, others) for d in _divisors(n))


def _fillna_and__numeric_cast(df: pd.DataFrame, value: float) -> pd.DataFrame:
    dtypes = df.dtypes
    for col in [c for c in df.columns if dtypes[c] == object]:
        vals = df[col]
        if all(_is_intlike(x) for x in vals):
            df[col] = vals.map(
                lambda x: (
                    value if (isinstance(x, float) and math.isnan(x)) else float(x)
                )
            ).astype(int)
        elif all(_is_numeric(x) for x in vals):
            df[col] = vals.map(
                lambda x: (
                    value if (isinstance(x, float) and math.isnan(x)) else float(x)
                )
            ).astype(float)

    cols = df.select_dtypes(include=[np.floating, float, np.integer, int]).columns
    df[cols] = df[cols].fillna(value)
    for col in df.columns:
        assert (
            not df[col].isna().any()
        ), f"df has nans in column {col} with dtype {df[col].dtype}"
    return df


def _numeric_cast(df: pd.DataFrame) -> pd.DataFrame:
    dtypes = df.dtypes
    for col in [c for c in df.columns if dtypes[c] == object]:
        series = df[col]
        if all(_is_intlike(x) for x in series):
            df[col] = series.map(float).astype(int)
        elif all(_is_numeric(x) for x in series):
            df[col] = series.map(float).astype(NUMPY_FLOAT_TYPE)
    return df



import os
import sys

sys.path.insert(0, os.path.abspath("_ext"))
sys.path.insert(0, os.path.abspath("../.."))  # Make your repo importable

import locale

locale.setlocale(locale.LC_ALL, "C.UTF-8")

# -- Project information -----------------------------------------------------
project = "accelforge"
author = "Tanner Andrulis, Michael Gilbert"
release = "0.1.0"

# -- HTML output -------------------------------------------------------------
# html_theme = 'sphinx_rtd_theme'
# html_theme = 'furo'
html_theme = "pydata_sphinx_theme"
# html_theme = 'sphinx_book_theme'
# pip3 install sphinx-furo-theme

html_static_path = ["_static"]
html_logo = "_static/logo.svg"

extensions = [
    "sphinx.ext.autodoc",  # Pull docstrings
    "sphinx.ext.napoleon",  # NumPy / Google style docstrings
    "sphinx.ext.autosummary",  # Generate autodoc summaries
    "sphinx.ext.viewcode",  # Add links to source code
    "sphinx_autodoc_typehints",  # Include type hints
    "sphinx.ext.intersphinx",  # Link to other projects' documentation
    "include_docstring",  # Include docstrings
    "include_notebook",  # Include notebooks
    "include_attrs",  # Include attributes & their docstrings
    "include_functions",  # Include functions & their docstrings
    "inherited_attributes",  # Inherit docstrings from parent classes
    "pydantic_defaults",  # Show default values for Pydantic fields
    "include_yaml",  # Include subsets of YAML files
    "sphinx_copybutton",  # Add copy button to code blocks
]


autodoc_default_options = {
    "members": True,
    "undoc-members": False,
    "exclude-members": "model_config,model_fields,__pydantic_fields__,model_post_init",
}

# Sphinx 9 introduced a new non-class-based autodoc directive that is the
# default. The custom InheritedAttributesClassDocumenter in _ext/ subclasses
# the legacy ClassDocumenter, so we keep the legacy path active so it runs.
autodoc_use_legacy_class_based = True

# ---------- Autodoc settings ----------
# Show type hints inline in signatures
autodoc_typehints = "signature"
autodoc_typehints_format = "short"

# Preserve default values
autodoc_preserve_defaults = True

# Force multi-line for long constructor signatures (Sphinx 7+)
autodoc_class_signature = "separated"

# ---------- HTML CSS to wrap signatures ----------
# Create docs/source/_static/custom.css with:
# .signature {
#     white-space: pre-wrap !important;
#     word-break: break-word;
# }
# html_static_path = ["_static"]
# html_css_files = ["custom.css"]
# html_js_files = ["custom.js"]

# ---------- Optional: Napoleon settings ----------
# If using Google/NumPy style docstrings
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_use_ivar = True

# Suppress warnings for things we can't control
suppress_warnings = [
    "ref.python",  # Multiple targets for same name
    # 'toc.not_included',  # Documents not in toctree (intentional)
    # 'misc.highlighting_failure',  # YAML lexing issues with Jinja2
]

nitpicky = True

# Suppress warnings for types that are intentionally private or external
nitpick_ignore = [
    # Private internal types (intentionally not documented)
    ("py:class", "accelforge.util._basetypes.EvalableModel"),
    ("py:class", "accelforge.util._basetypes.EvalableList"),
    ("py:class", "accelforge.util._basetypes.EvalableDict"),
    ("py:class", "accelforge.util._basetypes.ParsableDict"),
    ("py:class", "accelforge.util._basetypes.EvalsTo"),
    ("py:class", "accelforge.util._basetypes.TryEvalTo"),
    ("py:class", "accelforge.util._basetypes.EvalExtras"),
    ("py:class", "accelforge.util._basetypes.NoParse"),
    ("py:class", "accelforge.util._basetypes._get_tag"),
    ("py:class", "accelforge.util._setexpressions.InvertibleSet"),
    ("py:class", "accelforge.util._setexpressions.InvertibleSet[str]"),
    ("py:class", "accelforge.frontend.arch._ExtraAttrs"),
    ("py:class", "accelforge.util.parallel._SVGJupyterRender"),
    # Pydantic internal types (short name from type hints)
    ("py:class", "Tag"),
    ("py:class", "pydantic.types.Tag"),
    ("py:class", "pydantic.types.Discriminator"),
    ("py:class", "Discriminator"),
    ("py:class", "_get_tag"),
    ("py:class", "_Parallel"),
    # Short names from type hints (EvalsTo, sympy, etc.)
    ("py:class", "EvalsTo"),
    # Bogus refs from type hint / default rendering
    ("py:class", "None = None"),
    ("py:class", "TypeAlias MappingNodeTypes"),
    ("py:class", "TypeAlias NodeList"),
    # Internal mapper types (not in api docs)
    ("py:class", "accelforge.mapper.FFM._make_pmappings.pmapper_job.Job"),
    ("py:class", "accelforge.mapper.FFM._join_pmappings.compatibility.Compatibility"),
    # Method on private base (referenced from guide)
    ("py:meth", "accelforge.util._basetypes._FromYAMLAble.from_yaml"),
    # Python built-ins and standard library
    ("py:func", "math.pi"),
    ("py:func", "math.e"),
    ("py:func", "math.tau"),
    ("py:func", "math.inf"),
    ("py:func", "math.nan"),
    ("py:func", "range"),
    ("py:func", "float"),
    ("py:func", "float"),
    ("py:func", "int"),
    ("py:func", "str"),
    ("py:func", "bool"),
    ("py:func", "list"),
    ("py:func", "tuple"),
    # External library internals
    # Collections.abc inherited methods (auto-generated, not controllable)
    ("py:class", "D[k] if k in D, else d.  d defaults to None."),
    ("py:class", "a set-like object providing a view on D's items"),
    ("py:class", "a set-like object providing a view on D's keys"),
    ("py:class", "an object providing a view on D's values"),
    # Short type names from type hints (sphinx-autodoc-typehints emits short names;
    # targets are accelforge.frontend.renames.* and hwcomponents.ComponentModel)
    ("py:class", "IncEx"),
    ("py:class", "optional"),
    # Ideally these should be fixed, but idk how
    ("py:class", "EinsumName"),
    ("py:class", "Rank"),
    ("py:class", "RankVariable"),
    ("py:class", "TensorName"),
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    # 'matplotlib': ('https://matplotlib.org/stable/contents.html', None),
    # 'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    "scikit-learn": ("https://scikit-learn.org/stable/", None),
    "sympy": ("https://docs.sympy.org/latest/", None),
    "hwcomponents": ("https://accelergy-project.github.io/hwcomponents/", None),
}



import logging
from math import prod

from typing import Callable, Optional
import uuid
import copy

from joblib import delayed
from tqdm import tqdm


from accelforge.frontend import arch
from accelforge.frontend._workload_isl._symbolic import get_stride_and_halo
from accelforge.frontend.renames import Rank, RankVariable
from accelforge.frontend.spec import Spec
from accelforge.frontend.mapping import Loop, Mapping, TensorHolder
from accelforge.util._frozenset import oset
from accelforge.frontend._workload_isl._isl import (
    get_rank_variable_bounds,
    get_tensor_size,
    get_operation_space_size,
)
from accelforge.frontend.workload import EinsumName, SymbolTable, TensorName, Workload

from accelforge.mapper.FFM._make_pmappings.make_pmapping_templates import (
    make_pmapping_templates,
)
from accelforge.frontend.mapper.metrics import Metrics
from accelforge.mapper.FFM._make_pmappings.make_pmappings_from_templates import (
    make_pmappings_from_templates,
)
from accelforge.mapper.FFM._join_pmappings.compatibility import Compatibility
from accelforge.mapper.FFM._join_pmappings.pmapping_group import PmappingGroup
from accelforge.mapper.FFM._make_pmappings.make_pmappings_from_templates.symbol_relations import (
    get_initial_delta_choices,
)
from accelforge.util.parallel import (
    parallel,
    _memmap_read,
    get_n_parallel_jobs,
    is_using_parallel_processing,
)
from accelforge.mapper.FFM._make_pmappings.pmapper_job import (
    Job,
    SameCompatibilityJobs,
)


def get_rank_variable_bounds_for_all_einsums(spec: Spec):
    rank_variable_bounds = {
        einsum_name: get_rank_variable_bounds(spec.workload, einsum_name)
        for einsum_name in spec.workload.einsum_names
    }
    result = {}
    for e1, rv1 in rank_variable_bounds.items():
        result.update(rv1)
        for e2, rv2 in rank_variable_bounds.items():
            for r in oset(rv1.keys()) & oset(rv2.keys()):
                if rv1[r] != rv2[r]:
                    raise ValueError(
                        f"Rank variable {r} has different bounds for "
                        f"einsum {e1} and {e2}: {rv1[r]} and {rv2[r]}"
                    )
    return result


def get_n_computes(spec: Spec, einsum_name: EinsumName | None = None) -> int:
    einsums = spec.workload.einsums
    einsums = [einsum_name] if einsum_name is not None else spec.workload.einsum_names
    return sum(get_operation_space_size(spec.workload, e) for e in einsums)


def get_per_tensor_size(
    workload: Workload, return_n_elements: bool = False
) -> dict[TensorName, int]:
    sizes = {}
    for einsum in workload.einsums:
        for tensor in einsum.tensor_names:
            sizes[tensor] = get_tensor_size(workload, tensor)
            if not return_n_elements:
                sizes[tensor] *= einsum.tensor_accesses[tensor].bits_per_value
    return sizes


def get_jobs(
    spec: Spec,
    metrics: Metrics,
    einsum_names: list[EinsumName],
    fail_if_no_pmappings_for_einsum: bool,
    print_progress: bool,
) -> dict[EinsumName, dict[Compatibility, SameCompatibilityJobs]]:

    spec = spec

    einsum2jobs = {}
    fusable_tensors = spec.workload.tensor_names_used_in_multiple_einsums
    rank_variable_bounds = get_rank_variable_bounds_for_all_einsums(spec)

    einsum2spec: dict[EinsumName, Spec] = {}
    s = "Getting energy, latency, and leak power for components running each Einsum. "

    def _get_per_einsum_spec(spec, einsum_name):
        result = (
            spec._spec_eval_expressions(
                einsum_name=einsum_name,
                eval_arch=True,
                eval_non_arch=False,
            )
            .calculate_component_area_energy_latency_leak(
                einsum_name=einsum_name,
                area=False,
            )
            ._for_einsum(einsum_name)
            ._clear_component_models()
        )
        return einsum_name, _memmap_read(result)

    for einsum_name, result in parallel(
        [
            delayed(_get_per_einsum_spec)(spec, einsum_name)
            for einsum_name in einsum_names
        ],
        pbar=s if print_progress else None,
    ):
        einsum2spec[einsum_name] = result

    def make_jobs_for_einsum(
        einsum_name: EinsumName,
        spec: Spec,
        stride_and_halo: dict[
            EinsumName,
            dict[TensorName, dict[tuple[Rank, RankVariable], tuple[int, int]]],
        ],
        initial_delta_choices_for_einsum: dict[RankVariable, frozenset[int]],
        n_einsums: int,
        intermediate_tensors: set[TensorName],
    ):
        jobs = {}
        workload_einsum = spec.workload.einsums[einsum_name]
        for flattened_arch in spec._get_flattened_architecture(einsum_name=einsum_name):
            # Create jobs for each Einsum
            job = Job(
                spec_one_einsum=spec,
                einsum_name=einsum_name,
                metrics=metrics,
                rank_variable_bounds=rank_variable_bounds,
                flattened_arch=_memmap_read(flattened_arch),
                job_id=uuid.uuid4(),
                fusable_tensors=fusable_tensors & workload_einsum.tensor_names,
                stride_and_halo=stride_and_halo,
                initial_delta_choices=initial_delta_choices_for_einsum,
                objective_tolerance=spec.mapper.objective_tolerance,
                resource_usage_tolerance=spec.mapper.resource_usage_tolerance,
                workload_n_einsums=n_einsums,
                intermediate_tensors=intermediate_tensors
                & workload_einsum.tensor_names,
                explore_imperfect_spatial_loops=bool(
                    spec.mapper.explore_imperfect_spatial_loops
                ),
                explore_imperfect_temporal_loops=bool(
                    spec.mapper.explore_imperfect_temporal_loops
                ),
            )
            for j in make_pmapping_templates(job, print_progress):
                jobs.setdefault(j.compatibility, SameCompatibilityJobs()).append(j)

        return einsum_name, jobs

    stride_and_halo = get_stride_and_halo(spec.workload)
    initial_delta_chocies = {
        e: get_initial_delta_choices(e, spec.workload)
        for e in spec.workload.einsum_names
    }

    input_tensors = oset.union(*[e.input_tensor_names for e in spec.workload.einsums])
    output_tensors = oset.union(*[e.output_tensor_names for e in spec.workload.einsums])
    intermediate_tensors = input_tensors & output_tensors

    n_einsums = len(spec.workload.einsum_names)
    for einsum_name, jobs in parallel(
        [
            delayed(make_jobs_for_einsum)(
                einsum_name,
                spec._for_einsum(einsum_name),
                stride_and_halo,
                initial_delta_chocies[einsum_name],
                n_einsums,
                intermediate_tensors,
            )
            for einsum_name, spec in einsum2spec.items()
        ],
        pbar="Generating jobs" if print_progress else None,
    ):
        einsum2jobs.setdefault(einsum_name, {})
        for compatibility, job_list in jobs.items():
            einsum2jobs[einsum_name].setdefault(
                compatibility, SameCompatibilityJobs()
            ).extend(job_list)

    if fail_if_no_pmappings_for_einsum:
        for einsum_name, jobs in einsum2jobs.items():
            if len(jobs) == 0:
                raise ValueError(
                    f"No pmappings for {einsum_name}. Was the mapspace overconstrained?"
                )

    total_jobs = sum(len(jobs) for jobs in einsum2jobs.values())
    n_procs = get_n_parallel_jobs()
    memory_limit = min(
        spec.mapper.memory_limit, spec.mapper.memory_limit_per_process / n_procs
    )
    time_limit = min(
        spec.mapper.time_limit * n_procs / max(total_jobs, 1),
        spec.mapper.time_limit_per_pmapping_template,
    )
    total_templates = 0
    for einsum_name, compatibility_jobs in einsum2jobs.items():
        n_compatibility_jobs = sum(len(j) for j in compatibility_jobs.values())
        total_templates += n_compatibility_jobs
        if print_progress:
            s = "jobs" if n_compatibility_jobs > 1 else "job"
            print(f"Einsum {einsum_name} has {n_compatibility_jobs} pmapping {s}:")
        i = 0
        for job_list in compatibility_jobs.values():
            for job in job_list:
                if print_progress:
                    print(f"\t{i}\t{job.mapping.compact_str()}")
                    i += 1
                job.memory_limit = memory_limit
                job.time_limit = time_limit

    if print_progress:
        s = "templates" if total_templates > 1 else "template"
        print(f"Total number of pmapping {s}: {total_templates}")

    return einsum2jobs


def get_memories_to_track(
    spec: Spec,
    einsum2jobs: dict[EinsumName, list[Job]],
    metrics: Metrics,
    can_combine_multiple_runs: bool,
    print_progress: bool = True,
) -> tuple[list[str], list[str]]:

    memories_track_all = oset()
    for einsum, jobs in einsum2jobs.items():
        for job in jobs:
            memories_track_all.update(
                m.name for m in job.flattened_arch if isinstance(m, arch.Memory)
            )

    memories_track_pmappings_only = []
    ignored_resources = oset()

    # If we're combining the pmappings from multiple runs, we can't conclude anything
    # about the metrics to track
    if can_combine_multiple_runs or metrics.RESOURCE_USAGE in metrics:
        ignored_resources = oset()
        return (
            memories_track_all,
            memories_track_pmappings_only,
            ignored_resources,
        )

    tensor_sizes = {}
    for tensor, size in get_per_tensor_size(spec.workload).items():
        scale = 1
        for einsum in spec.workload.einsums_with_tensor(tensor):
            if einsum.tensor_accesses[tensor].persistent:
                scale = max(scale, spec.workload.n_instances * einsum.n_instances)
        tensor_sizes[tensor] = size * scale

    # If the memory is big enough to hold all the tensors then we don't need to consider
    # it
    for memory in list(memories_track_all):
        usage = 0
        for einsum in einsum2jobs.keys():
            job = einsum2jobs[einsum][0]
            try:
                mem: arch.Memory = job.spec_one_einsum.arch.find(memory)
            except ValueError:
                continue
            for tensor in spec.workload.einsums[einsum].tensor_names:
                if mem.size == 0:
                    usage = 2  # FAIL
                else:
                    workload_bpv = (
                        spec.workload.einsums[einsum]
                        .tensor_accesses[tensor]
                        .bits_per_value
                    )
                    effective_bpv = mem.bits_per_value.get(tensor, workload_bpv)
                    usage += (
                        tensor_sizes[tensor] * effective_bpv / (workload_bpv * mem.size)
                    )

        if usage <= 1:
            ignored_resources.add(memory)
            if print_progress:
                print(
                    f"Not tracking memory {memory}. It is big enough to simultaneously "
                    f"hold every workload tensor that may be stored in it. Max "
                    f"possible usage: {usage * 100:.2f}%"
                )
            memories_track_all.remove(memory)

    # If the memory is below every backing tensor holder node, then we need it for the
    # pmapping exploration but can drop it immediately
    for m in list(memories_track_all):
        must_track = False
        for _, jobs in einsum2jobs.items():
            for job in jobs:
                seen = False
                for node in job.mapping.nodes:
                    if isinstance(node, TensorHolder) and node.component == m:
                        seen = True
                        if node.persistent:
                            ignored_resources.add(m)
                        if node._backing:
                            must_track = True
                    if isinstance(node, Loop) and node._fused and seen:
                        must_track = True

        if not must_track:
            memories_track_all.remove(m)
            memories_track_pmappings_only.append(m)
            if print_progress:
                print(
                    f"Not tracking memory {m} across joining stages. It is never "
                    f"reserved across fused loop iterations."
                )

    return memories_track_all, memories_track_pmappings_only, ignored_resources


def make_pmappings(
    spec: Spec,
    can_combine_multiple_runs: bool,
    metrics: Metrics = Metrics.ENERGY | Metrics.LATENCY,
    einsum_names: Optional[list[EinsumName]] = None,
    fail_if_no_pmappings_for_einsum: bool | None = None,
    print_progress: bool = True,
    one_pbar_only: bool = False,
) -> tuple[
    dict[EinsumName, list[PmappingGroup]],
    dict[EinsumName, dict[uuid.UUID, Mapping]],
    dict[EinsumName, list[Job]],
]:
    """
    Explores pmapspace of `einsum_names` (default: all Einsums in workload).
    """
    spec = copy.deepcopy(spec)

    if einsum_names is None:
        einsum_names = spec.workload.einsum_names

    if one_pbar_only:
        assert (
            not print_progress
        ), "print_progress must be False if one_pbar_only is True"

    if fail_if_no_pmappings_for_einsum is None:
        fail_if_no_pmappings_for_einsum = not can_combine_multiple_runs

    spec = spec._spec_eval_expressions(
        eval_arch=False,
        eval_non_arch=True,
    )

    einsum2jobs = {}
    new_einsum2jobs = get_jobs(
        spec,
        metrics,
        einsum_names,
        fail_if_no_pmappings_for_einsum,
        print_progress,
    )
    _fill_jobs_with_memories_to_track(
        new_einsum2jobs, spec, metrics, can_combine_multiple_runs, print_progress
    )
    for einsum_name, jobs in new_einsum2jobs.items():
        einsum2jobs.setdefault(einsum_name, {})
        for compatibility, job_list in jobs.items():
            einsum2jobs[einsum_name].setdefault(
                compatibility, SameCompatibilityJobs()
            ).extend(job_list)

    calls = _allocate_jobs(einsum2jobs, print_progress)

    # Sort the calls by the length of the longest mapping in each job. We get long
    # poles with the long mappings, so we want to get them done early so we don't
    # have one or two procs slowing us down at the end.
    def get_longest_mapping_length(call):
        j: SameCompatibilityJobs = call[2]["jobs_with_similar_compatibilities"]
        return max([len(j2.mapping.nodes) for j2 in j])

    calls = sorted(calls, key=get_longest_mapping_length, reverse=True)

    pmapping_objects = {}
    pmapping_groups = {einsum_name: [] for einsum_name in spec.workload.einsum_names}
    keep_rates = []
    for (
        einsum_name,
        new_pmapping_groups,
        pmappings,
        cur_keep_rates,
    ) in parallel(
        calls,
        pbar=f"Generating pmappings" if print_progress or one_pbar_only else None,
        return_as="generator_unordered",
    ):
        pmapping_groups[einsum_name].extend(new_pmapping_groups)
        pmapping_objects.setdefault(einsum_name, {}).update(pmappings)
        keep_rates.extend(cur_keep_rates)

    for einsum_name in list(pmapping_groups.keys()):
        pmapping_groups[einsum_name] = PmappingGroup.combine_combineable(
            pmapping_groups[einsum_name],
            "All",
            pbar_postfix=f" for {einsum_name}",
            print_progress=print_progress,
        )

    return_jobs = {
        e: [j for jobs in v.values() for j in jobs] for e, v in einsum2jobs.items()
    }

    # Propagate keep_rates from worker-side job copies back to original jobs
    id2keep_rates = {jid: (keep_rates, n) for jid, keep_rates, n in keep_rates}
    for jobs in return_jobs.values():
        for job in jobs:
            assert job.job_id in id2keep_rates, f"Job {job.job_id} not found"
            job.pmapping_keep_rates.update(id2keep_rates[job.job_id][0])
            job.n_total_pmappings = id2keep_rates[job.job_id][1]

    return pmapping_groups, pmapping_objects, return_jobs


def _raise_error_if_no_pmappings(einsum2jobs):
    for einsum_name, jobs in einsum2jobs.items():
        if len(jobs) == 0:
            raise ValueError(
                f"No pmappings for {einsum_name}. " f"Was the mapspace overconstrained?"
            )


def _allocate_jobs(einsum2jobs, print_progress: bool = True):
    calls = []
    for einsum_name, jobs in einsum2jobs.items():
        calls.extend(
            delayed(make_pmappings_from_templates)(
                jobs_with_similar_compatibilities=job_list,
            )
            for job_list in jobs.values()
        )

    split = True

    if split:
        calls = []
        for einsum_name, jobs in einsum2jobs.items():
            for job_list in jobs.values():
                calls.extend(
                    delayed(make_pmappings_from_templates)(
                        jobs_with_similar_compatibilities=job,
                    )
                    for job in job_list.split()
                )
    return calls


def _fill_jobs_with_memories_to_track(
    einsum2jobs: dict[EinsumName, dict[Compatibility, SameCompatibilityJobs]],
    spec,
    metrics,
    can_combine_multiple_runs,
    print_progress,
):
    einsum2jobs_flattened = {
        e: [j for jobs in v.values() for j in jobs] for e, v in einsum2jobs.items()
    }

    memories_track_all, memories_track_pmappings_only, ignored_resources = (
        get_memories_to_track(
            spec,
            einsum2jobs_flattened,
            metrics,
            can_combine_multiple_runs,
            print_progress,
        )
    )
    for jobs in einsum2jobs_flattened.values():
        for j in jobs:
            j.memories_track_all = memories_track_all
            j.memories_track_pmappings_only = memories_track_pmappings_only
            j.ignored_resources = ignored_resources



from collections import defaultdict
import logging
from typing import List
from accelforge._accelerated_imports import np
from accelforge.frontend._workload_isl._symbolic import PartiallyRelevant, Relevant
import accelforge.frontend.arch as arch
from accelforge.frontend.arch import (
    Comparison,
    _MinUsageConstraintLambda,
    _TileShapeConstraintLambda,
    _LoopBoundsConstraintLambda,
    _ConstraintLambda,
)
from accelforge.frontend.mapping import (
    Loop,
    MappingNode,
    TensorHolder,
    Temporal,
    Spatial,
)
from accelforge.frontend.renames import TensorName
from accelforge.frontend.workload import EinsumName, RankVariable
from accelforge.util._setexpressions import InvertibleSet
from accelforge.util._frozenset import fzs, oset


# =================================================================================================
# Attach constraints to mapping
# =================================================================================================
class MappingConstraints:
    def __init__(self):
        self.tile_shape_constraints: list[_TileShapeConstraintLambda] = []
        self.loop_bounds_constraints: list[_LoopBoundsConstraintLambda] = []
        self.min_usage_constraints: dict[tuple[str, str], _MinUsageConstraintLambda] = (
            {}
        )

    def get_all_constraints(self) -> list[_ConstraintLambda]:
        return (
            self.tile_shape_constraints
            + self.loop_bounds_constraints
            + list(self.min_usage_constraints.values())
        )

    def set_loop_indices(self, nodes: list[MappingNode]):
        loops = [n for n in nodes if isinstance(n, Loop)]
        for c in self.get_all_constraints():
            c._target_node_indices = [nodes.index(t) for t in c.target_mapping_nodes]
            c._target_loop_indices = [loops.index(t) for t in c.target_mapping_nodes]

        # Min usage constraints also depend on the loop ABOVE the target loop
        # because the loop above determines the number of tiles
        for c in self.min_usage_constraints.values():
            # Rank variables must be unique between mapping nodes
            rank_variables = oset(t.rank_variable for t in c.target_mapping_nodes)
            assert len(rank_variables) == len(
                c.target_mapping_nodes
            ), "Rank variables must be unique between mapping nodes"

            for target_mapping_node in c.target_mapping_nodes:
                assert isinstance(target_mapping_node, Spatial)
                loop_index = loops.index(target_mapping_node) - 1
                while loop_index >= 0:
                    loop = loops[loop_index]
                    if loop.rank_variable in rank_variables:
                        c._target_loop_indices.append(loop_index)
                        c._target_node_indices.append(nodes.index(loop))
                        break
                    loop_index -= 1

    def clear_constrained_to_one(
        self, mapping: list["MappingNode"], einsum_name: EinsumName
    ) -> list["MappingNode"]:
        # Not constrained to one --> Can't remove
        node2constraints = defaultdict(list)
        do_not_remove = oset()
        for c in self.tile_shape_constraints:
            for t in c.target_mapping_nodes:
                node2constraints[id(t)].append(c)
                do_not_remove.add(id(t))
        for c in self.loop_bounds_constraints:
            if not c.constraint._constrained_to_one():
                for t in c.target_mapping_nodes:
                    node2constraints[id(t)].append(c)
                    do_not_remove.add(id(t))

        # Constrained to one --> remove iff not in do_not_remove
        to_remove = oset()
        for c in self.loop_bounds_constraints:
            if c.constraint._constrained_to_one():
                my_remove = oset(id(t) for t in c.target_mapping_nodes)
                if my_remove & do_not_remove:
                    loops = [n for n in mapping if id(n) in my_remove]
                    p = len(loops) == 1
                    loops = (", ".join(n.compact_str() for n in loops)).strip()
                    isare = "is" if p else "are"
                    all_others = ", ".join(
                        str(c2) for c2 in node2constraints[id(t)] if c2 != c
                    )
                    logging.warning(
                        f"For Einsum {einsum_name}, loop{'s' * (not p)} {loops} "
                        f"{isare} set to be removed by {c} and also appear{'s' * p} in "
                        f"{all_others}. The loop{'s' * (not p)} will not be removed "
                        f"from the mapping, but it may be subject to conflicting "
                        f"constraints."
                    )

                c.target_mapping_nodes = [
                    t for t in c.target_mapping_nodes if id(t) not in my_remove
                ]
                to_remove.update(my_remove)
        self.loop_bounds_constraints = [
            c
            for c in self.loop_bounds_constraints
            if not c.constraint._constrained_to_one()
        ]

        for c in self.get_all_constraints():
            c.target_mapping_nodes = [
                n for n in c.target_mapping_nodes if id(n) not in to_remove
            ]

        return [m for m in mapping if id(m) not in to_remove]

    def pretty_str(self) -> str:
        s = ""
        all_constraints = self.get_all_constraints()
        s += "Tile shape constraints:\n"
        for c in self.tile_shape_constraints:
            s += f"\t{all_constraints.index(c)} {c.pretty_str()}\n"
        s += "Loop bounds constraints:\n"
        for c in self.loop_bounds_constraints:
            s += f"\t{all_constraints.index(c)} {c.pretty_str()}\n"
        s += "Min usage constraints:\n"
        for c in self.min_usage_constraints.values():
            s += f"\t{all_constraints.index(c)} {c.pretty_str()}\n"
        return s

    def remove_missing_targets(self, mapping: list[MappingNode]):
        for c in self.get_all_constraints():
            c.target_mapping_nodes = [n for n in c.target_mapping_nodes if n in mapping]

        self.tile_shape_constraints = [c for c in self.tile_shape_constraints if c]
        self.loop_bounds_constraints = [c for c in self.loop_bounds_constraints if c]
        self.min_usage_constraints = {
            k: c for k, c in self.min_usage_constraints.items() if c
        }


def first_tensor_holder_index(mapping: list["MappingNode"], memory_name: str) -> int:
    for i, m in enumerate(mapping):
        if isinstance(m, TensorHolder) and m.component == memory_name:
            return i
    return None


def constrained_loops(
    mapping: list["MappingNode"],
    rank_variables: set[RankVariable],
    start_index: int = None,
    look_behind: bool = False,
    component: str = None,
    one_loop_per_rank_variable: bool = True,
) -> list[Loop]:
    """The first loop nodes containing given rank variables."""
    nodes = []
    remaining_rank_variables = oset(rank_variables)

    if look_behind:
        to_check = list(enumerate(mapping))
        to_check.reverse()
        if start_index is not None:
            to_check = [m for i, m in to_check if i <= start_index]
    else:
        to_check = list(enumerate(mapping))
        to_check = [m for i, m in to_check if start_index is None or i >= start_index]

    for m in to_check:
        if not isinstance(m, Loop):
            continue
        if component is not None and (
            not isinstance(m, Spatial) or m.component != component
        ):
            continue
        assert isinstance(m.rank_variable, RankVariable)
        if m.rank_variable in remaining_rank_variables:
            nodes.append(m)
            if one_loop_per_rank_variable:
                remaining_rank_variables.discard(m.rank_variable)
    # TODO: what is this supposed to do?
    # for r in remaining_rank_variables:
    #     assert (
    #         component is None
    #     ), "There should be a spatial loop for every rank variable"
    return nodes


def get_constraints(
    flattened_arch: list[arch.Leaf],
    mapping: List[MappingNode],
    symbol_table: dict[str, InvertibleSet],
    einsum_name: EinsumName,
    tensor_to_relevancy: dict[
        TensorName, dict[RankVariable, Relevant | PartiallyRelevant]
    ],
    is_copy_operation: bool,
) -> tuple[List[MappingNode], MappingConstraints]:

    constraints = MappingConstraints()

    # Tensor constraints
    for m in flattened_arch:
        # Ignore if not a memory
        if not isinstance(m, arch.Memory):
            continue

        # Ignore if it doesn't hold any tensors
        if (index := first_tensor_holder_index(mapping, m.name)) is None:
            continue

        # Tile shape constraints
        for c in m.tensors.tile_shape:
            nodes = constrained_loops(
                mapping, c.expression, index - 1, look_behind=True
            )
            for exp in c._split_expression():
                new_nodes = [n for n in nodes if n.rank_variable in exp]
                constraint = _TileShapeConstraintLambda(c, new_nodes, exp)
                constraints.tile_shape_constraints.append(constraint)

        no_resend_refetch_constraint = (
            _loop_bound_constraint_from_no_refetch_and_resend(
                mapping, index, m, symbol_table, tensor_to_relevancy
            )
        )
        if no_resend_refetch_constraint is not None:
            constraints.loop_bounds_constraints.append(no_resend_refetch_constraint)

    # Spatial constraints
    for m in flattened_arch:
        if not isinstance(m, arch.Leaf):
            continue

        for dim in m.spatial:
            loops = [
                n
                for n in mapping
                if isinstance(n, Spatial)
                and (n.component, n.name) == (m.name, dim.name)
            ]
            loop_bounds = list(dim.loop_bounds)
            if dim.reuse:
                loop_bounds.append(
                    Comparison(
                        expression=dim.reuse.rank_variables,
                        operator="==",
                        value=1,
                    )
                )
                loop_bounds[-1]._str_repr = f"reuse {oset(dim.reuse)}"

            # Loop bounds constraints
            if loop_bounds:
                for c in loop_bounds:
                    nodes = constrained_loops(loops, c.expression, component=m.name)
                    for exp in c._split_expression():
                        new_nodes = [l for l in loops if l.rank_variable in exp]
                        constraint = _LoopBoundsConstraintLambda(c, new_nodes, exp)
                        constraints.loop_bounds_constraints.append(constraint)

            # Min usage constraints
            target_mapping_nodes = [
                n
                for n in mapping
                if isinstance(n, Spatial)
                and n.component == m.name
                and n.name == dim.name
            ]
            # Min usage constraints don't apply to copy operations
            if dim.min_usage > 0 and not is_copy_operation:
                if not target_mapping_nodes:
                    continue
                rank_variables = oset(t.rank_variable for t in target_mapping_nodes)
                constraint = _MinUsageConstraintLambda(
                    target_mapping_nodes,
                    rank_variables,
                    dim.min_usage,
                )
                key = (m.name, dim.name)
                constraints.min_usage_constraints[key] = constraint

            for t in target_mapping_nodes:
                t._may_reuse = dim.may_reuse

    # Additional spatial constraints
    for m in mapping:
        if isinstance(m, Spatial) and m._constrained_to_one:
            constraints.loop_bounds_constraints.append(
                _LoopBoundsConstraintLambda(
                    Comparison(expression=m.rank_variable, operator="==", value=1),
                    [m],
                    m.rank_variable,
                )
            )

    mapping = constraints.clear_constrained_to_one(mapping, einsum_name)

    return mapping, constraints


def _loop_bound_constraint_from_no_refetch_and_resend(
    mapping, index, arch_node, symbol_table, tensor_to_relevancy
):
    """Create a loop bound constraint representing no refetch and no resend."""
    no_refetch = mapping[index].component_object.tensors.no_refetch_from_above
    exp = symbol_table[arch_node.name] & no_refetch

    nodes = []
    seen = oset()
    for no_refetch in exp.iter_one_element_sets():
        # Start from the first index of the tensor holder, stop at index - 1
        start_index = 0
        n = next(iter(no_refetch))
        while start_index < len(mapping):
            if (
                isinstance(mapping[start_index], TensorHolder)
                and n in mapping[start_index].tensors
            ):
                break
            start_index += 1

        end_index = start_index + 1
        while end_index < len(mapping):
            if (
                isinstance(mapping[end_index], TensorHolder)
                and n in mapping[end_index].tensors
                and mapping[end_index].component == arch_node.name
            ):
                break
            end_index += 1
        else:
            # This tensor isn't stored in this component at all. Don't look at any
            # loops. We can also end up here if the tensor is backed in this
            # component, in which case we'll start looking below the component &
            # never find it.
            end_index = start_index

        for i in range(start_index, min(end_index, len(mapping))):
            if isinstance(mapping[i], Temporal) and not isinstance(
                tensor_to_relevancy[n][mapping[i].rank_variable], Relevant
            ):
                if id(mapping[i]) not in seen:
                    seen.add(id(mapping[i]))
                    nodes.append(mapping[i])

    no_resend = mapping[index].component_object.tensors.no_resend_to_below
    for no_resend in no_resend.iter_one_element_sets():
        # Start from the first index of this one, stop when we find someone else
        # below
        start_index = 0
        n = next(iter(no_resend))
        while start_index < len(mapping):
            if (
                isinstance(mapping[start_index], TensorHolder)
                and n in mapping[start_index].tensors
                and mapping[start_index].component == arch_node.name
            ):
                break
            start_index += 1

        end_index = start_index + 1
        while end_index < len(mapping):
            if (
                isinstance(mapping[end_index], TensorHolder)
                and n in mapping[end_index].tensors
            ):
                # Can't have two tensor holders for the same tensor + component
                assert mapping[end_index].component != arch_node.name
                break
            end_index += 1

        for i in range(start_index, min(end_index, len(mapping))):
            if isinstance(mapping[i], Temporal) and not isinstance(
                tensor_to_relevancy[n][mapping[i].rank_variable], Relevant
            ):
                if id(mapping[i]) not in seen:
                    seen.add(id(mapping[i]))
                    nodes.append(mapping[i])

    if nodes:
        return _LoopBoundsConstraintLambda(
            Comparison(expression=exp, operator="==", value=1), nodes, exp
        )
    else:
        return None



from sphinx.ext.autodoc import ClassDocumenter, INSTANCEATTR, ObjectMember
from sphinx.util.inspect import safe_getattr
import inspect
from pydantic import BaseModel

# Default ignore lists
DEFAULT_IGNORE = [
    "object",
    "ABC",
    "BaseModel",
    "Enum",
    "Flag",
    "IntEnum",
    "IntFlag",
    "Generic",
    "object",
    "ABC",
    "BaseModel",
    "EvalableModel",
    "EvalableList",
    "EvalableDict",
    "EvalExtras",
    "NonEvalableModel",
    "_FromYAMLAble",
    "EvalsTo",
    "Enum",
    "Flag",
    "IntEnum",
    "IntFlag",
    "Generic",
    "model_validate",
    "model_validate_json",
    "model_validate_strings",
    "parse_file",
    "parse_obj",
    "parse_raw",
    "schema",
    "schema_json",
    "model_dump",
    "model_dump_json",
    "model_copy",
    "construct",
    "copy",
    "dict",
    "json",
    "update_forward_refs",
    "model_post_init",
    "model_config",
    "model_fields",
    "model_fields_set",
    "model_extra",
    "model_private",
    "model_rebuild",
    "get_fields",
    "get_validator",
    "all_fields_default",
    "model_dump_non_none",
] + list(BaseModel.__dict__.keys())

# Methods that should appear on every documented subclass even when their
# defining class is in DEFAULT_IGNORE. These are part of the public API.
ALWAYS_INHERIT = ["to_yaml", "from_yaml"]


class InheritedAttributesClassDocumenter(ClassDocumenter):
    """Enhanced ClassDocumenter that includes inherited attributes."""

    priority = ClassDocumenter.priority + 1

    def get_object_members(self, want_all):
        """Override to include inherited members."""
        # Get the normal members first
        members_check_module, members = super().get_object_members(want_all)

        # Get configuration
        config = self.env.config
        ignore = getattr(config, "inherited_attributes_ignore", []) + DEFAULT_IGNORE

        # Collect ALL attributes defined directly on this class (not inherited)
        own_members = set()
        if hasattr(self.object, "__dict__"):
            own_members.update(self.object.__dict__.keys())
        if hasattr(self.object, "model_fields"):
            own_members.update(self.object.model_fields.keys())
        if hasattr(self.object, "__fields__"):
            own_members.update(self.object.__fields__.keys())

        # Filter out members from ignored classes that were already included
        # BUT keep them if they're overridden in the current class
        filtered_members = []
        for member in members:
            if hasattr(member, "__getitem__") and len(member) >= 4:
                name = member[0]
                member_class = member[3] if len(member) > 3 else None

                # Keep if it's defined directly on this class
                if name in own_members:
                    filtered_members.append(member)
                    continue

                # Skip if from an ignored class (and not overridden)
                if member_class and self._should_ignore_class(member_class, ignore):
                    continue

                # Skip ignored functions/attributes (only if not overridden)
                if name in ignore:
                    continue

                filtered_members.append(member)
            else:
                filtered_members.append(member)

        members = filtered_members

        # Get MRO
        try:
            mro = inspect.getmro(self.object)
        except (AttributeError, TypeError):
            return members_check_module, members

        # Track seen names - includes both members from super() and own_members
        # ObjectMember is a namedtuple, access by index
        seen_names = set()
        for member in members:
            if hasattr(member, "__getitem__"):
                seen_names.add(member[0])
        # Add all own members to seen_names to prevent inheritance
        seen_names.update(own_members)

        # Collect inherited members
        inherited = []

        for parent_class in mro[1:]:  # Skip self
            # Stop at the first ignored ancestor — we don't want to surface the
            # internals (get_fields, get_validator, etc.) of EvalableModel and
            # friends. Specific must-have methods are added separately below
            # via ALWAYS_INHERIT.
            if self._should_ignore_class(parent_class, ignore):
                break

            # Get members defined directly on this parent. Using vars() instead
            # of dir() means we attribute each method to the class that actually
            # defines it, so the break above correctly excludes everything from
            # ignored ancestors (e.g. _OurBaseModel.shallow_model_dump won't
            # leak in via Spatialable, which inherits but does not redefine it).
            for name in vars(parent_class):
                # Skip underscore-prefixed
                if name.startswith("_"):
                    continue

                # Skip if already seen (including own_members)
                if name in seen_names:
                    continue

                # Skip ignored
                if name in ignore:
                    continue

                try:
                    obj = safe_getattr(parent_class, name)

                    # Skip callables in ignore list
                    if callable(obj) and name in ignore:
                        continue

                    # Create ObjectMember
                    inherited.append(ObjectMember(name, obj))
                    seen_names.add(name)

                except (AttributeError, TypeError):
                    continue

            # Check Pydantic fields
            if hasattr(parent_class, "model_fields"):
                for field_name in parent_class.model_fields:
                    if field_name.startswith("_"):
                        continue
                    # Skip if already seen (including own_members)
                    if field_name in seen_names:
                        continue
                    if field_name in ignore:
                        continue

                    inherited.append(ObjectMember(field_name, INSTANCEATTR))
                    seen_names.add(field_name)

            elif hasattr(parent_class, "__fields__"):
                for field_name in parent_class.__fields__:
                    if field_name.startswith("_"):
                        continue
                    # Skip if already seen (including own_members)
                    if field_name in seen_names:
                        continue
                    if field_name in ignore:
                        continue

                    inherited.append(ObjectMember(field_name, INSTANCEATTR))
                    seen_names.add(field_name)

        # Force-include public API methods that live on ignored ancestors.
        # Use vars() to find the class that actually defines the method, so the
        # source link points at the original definition rather than a re-export.
        for method_name in ALWAYS_INHERIT:
            if method_name in seen_names:
                continue
            for parent_class in mro[1:]:
                if method_name in vars(parent_class):
                    try:
                        obj = safe_getattr(parent_class, method_name)
                    except (AttributeError, TypeError):
                        break
                    inherited.append(ObjectMember(method_name, obj))
                    seen_names.add(method_name)
                    break

        # Combine
        all_members = list(members) + inherited

        return members_check_module, all_members

    def _should_ignore_class(self, cls, ignore_classes):
        """Check if a class should be ignored."""
        name = getattr(cls, "__name__", "")
        module = getattr(cls, "__module__", "")

        # Ignore if not part of accelforge package
        if module and not module.startswith("accelforge"):
            return True

        # Check against ignore list using just the class name
        for ignore_pattern in ignore_classes:
            if name == ignore_pattern:
                return True

        return False


def setup(app):
    """Setup the extension."""
    # Only add config value if it doesn't exist
    if not hasattr(app.config, "inherited_attributes_ignore"):
        app.add_config_value("inherited_attributes_ignore", [], "env")

    # Sphinx 9 re-registers the default ClassDocumenter at the config-inited
    # event (see sphinx.ext.autodoc._register_directives), which clobbers any
    # autodocumenter we register here in setup(). Hook the same event with a
    # later priority to re-install our subclass after sphinx has settled.
    def _install(app, config):
        app.add_autodocumenter(InheritedAttributesClassDocumenter, override=True)

    app.connect("config-inited", _install, priority=1000)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Version scheme for setuptools-scm - creates post-release versions."""

from setuptools_scm.version import guess_next_version


def post_version(version):
    """Create post-release versions instead of dev versions."""
    if version.exact:
        return version.format_with("{tag}").lstrip("v")

    base = (
        str(version.tag).lstrip("v")
        if version.tag
        else (guess_next_version(version) or "1.0")
    )
    distance = version.distance or 0

    return f"{base}.{distance}" if distance > 0 else base


def no_local(version):
    """No local version identifier."""
    return ""



import copy
from dataclasses import dataclass, field
from typing import Any

from accelforge.frontend.mapping import Compute, Mapping
from accelforge.frontend.workload import TensorName

from accelforge.model._looptree.types import Buffet, Compute, Network

from accelforge.util._frozenset import oset
from accelforge.util._sympy.broadcast_max import (
    MaxGeqZero,
    MinGeqZero,
    min_nonzero,
    max_dict,
)

import sympy


@dataclass
class NetworkStats:
    total_hops: Any = field(default=0)
    max_hops: Any = field(default=0)

    def repeat(self, n_repeats):
        new = copy.copy(self)
        if n_repeats == 1:
            return new
        if type(n_repeats) is float and n_repeats == int(n_repeats):
            n_repeats = int(n_repeats)
        new.total_hops = new.total_hops * n_repeats
        return new

    def combine(self, other: "NetworkStats"):
        self.total_hops += other.total_hops
        self.max_hops = max(self.max_hops, other.max_hops)


@dataclass
class BuffetStats:
    total_reads_to_parent: Any = field(default=0)
    total_writes_to_parent: Any = field(default=0)
    max_per_parent_reads_to_parent: Any = field(default=0)
    max_per_parent_writes_to_parent: Any = field(default=0)

    total_reads_to_peer: Any = field(default=0)
    total_writes_to_peer: Any = field(default=0)
    max_per_unit_reads_to_peer: Any = field(default=0)
    max_per_unit_writes_to_peer: Any = field(default=0)

    # Skip the first iteration of temporal loops for data that is written
    total_skipped_first_reads_to_parent: Any = field(default=0)
    total_skipped_first_reads_to_peer: Any = field(default=0)
    min_per_parent_skipped_first_reads_to_parent: Any = field(default=0)
    min_per_unit_skipped_first_writes_to_peer: Any = field(default=0)

    max_occupancy: Any = field(default=0)
    _n_loops_above: int = field(default=0)

    # These are used to calculate energy and latency
    total_write_actions: Any = field(default=0)
    max_per_unit_write_actions: Any = field(default=0)
    total_read_actions: Any = field(default=0)
    max_per_unit_read_actions: Any = field(default=0)

    total_skipped_first_write_actions: Any = field(default=0)
    min_per_unit_skipped_first_write_actions: Any = field(default=0)
    total_skipped_first_read_actions: Any = field(default=0)
    min_per_unit_skipped_first_read_actions: Any = field(default=0)

    # NOTE: anything other than min_, max_, or total_ must default to
    # None. There are asserts that check this.
    persistent: bool = field(default=None)

    @property
    def n_loops_above(self) -> int:
        if self.persistent:
            return -1
        return self._n_loops_above

    @n_loops_above.setter
    def n_loops_above(self, value: int):
        self._n_loops_above = value

    def repeat_temporal(self, factor: int, is_fully_relevant: bool) -> "BuffetStats":
        new = copy.copy(self)
        if factor == 1:
            return new
        if type(factor) is float and factor == int(factor):
            factor = int(factor)  # sympy Symbol * int is 4× faster than * float
        for k, v in new.__dict__.items():
            if not k.startswith(("total_", "max_", "min_")):
                continue
            if "skipped_first" in k and not is_fully_relevant:
                continue  # First actions occur once per relevant iteration.
            if k == "max_occupancy":
                continue  # Max occupancy is not affected by temporal loops above
            new.__dict__[k] = v * factor
        return new

    def repeat_spatial(self, factor: int, reuse_parent_accesses: bool) -> "BuffetStats":
        new = copy.copy(self)
        if factor == 1:
            return new
        if type(factor) is float and factor == int(factor):
            factor = int(factor)
        for k, v in new.__dict__.items():
            if not k.startswith(("total_", "max_", "min_")):
                continue
            if "parent" in k and reuse_parent_accesses:
                continue  # If parent accesses are reused, no need to multiply
            if "per_unit" in k:
                continue  # Spatial fanout doesn't affect per-unit stats
            if k == "max_occupancy":
                continue  # Max occupancy is not affected by temporal loops above
            new.__dict__[k] = v * factor
        return new

    def max(self, **kwargs: Any):
        for key, value in kwargs.items():
            setattr(self, key, MaxGeqZero(getattr(self, key), value))

    def min(self, **kwargs: Any):
        for key, value in kwargs.items():
            setattr(self, key, MinGeqZero(getattr(self, key), value))

    def __add__(self, other: "BuffetStats") -> "BuffetStats":
        new = copy.copy(self)
        for k, v in self.__dict__.items():
            other_v = other.__dict__[k]
            if k.startswith("min_"):
                new.__dict__[k] = min_nonzero(v, other_v)
            elif k.startswith("max_"):
                new.__dict__[k] = MaxGeqZero(v, other_v)
            elif k.startswith("total_"):
                new.__dict__[k] = v + other_v
            elif v is None:
                new.__dict__[k] = other_v
            else:
                if v is None:
                    new.__dict__[k] = other_v
                else:
                    assert (
                        v == other_v
                    ), f"BUG: {k} is different. self: {v} other: {other_v}"
        return new

    def __iadd__(self, other: "BuffetStats") -> "BuffetStats":
        new = self + other
        for key, value in new.__dict__.items():
            setattr(self, key, value)
        return self

    def net_total_read_actions(self) -> Any:
        return self.total_read_actions - self.total_skipped_first_read_actions

    def net_total_write_actions(self) -> Any:
        return self.total_write_actions - self.total_skipped_first_write_actions

    def net_max_per_unit_read_actions(self) -> Any:
        return (
            self.max_per_unit_read_actions
            - self.min_per_unit_skipped_first_read_actions
        )

    def net_max_per_unit_write_actions(self) -> Any:
        return (
            self.max_per_unit_write_actions
            - self.min_per_unit_skipped_first_write_actions
        )

    @classmethod
    def blank(cls):
        stats = cls()
        stats.n_loops_above = None  # Inherit from whoever is added to this
        return stats


@dataclass
class ComputeStats:
    total_ops: Any = field(default=0)
    max_per_unit_ops: Any = field(default=0)
    # "max" below refers to the longest latency of any iteration
    max_latency: Any = field(default=0)
    # Mapping from the loop-index (0 at top) to the latency of the first
    # iteration of that loop. "Max" because we may have loops above that and we
    # will take the maximum of the firsts.
    max_first_latency: dict[int, Any] = field(default_factory=dict)

    def repeat_temporal(self, factor: int) -> "ComputeStats":
        new = copy.copy(self)
        if factor == 1:
            return new
        if type(factor) is float and factor == int(factor):
            factor = int(factor)
        new.total_ops = new.total_ops * factor
        new.max_per_unit_ops = new.max_per_unit_ops * factor
        new.max_latency = new.max_latency * factor
        # NOTE: max_first_latency does not change
        return new

    def repeat_spatial(self, factor: int) -> "ComputeStats":
        new = copy.copy(self)
        if factor == 1:
            return new
        if type(factor) is float and factor == int(factor):
            factor = int(factor)
        new.total_ops = new.total_ops * factor
        return new

    def __add__(self, other: "ComputeStats") -> "ComputeStats":
        new = copy.copy(self)
        new.total_ops += other.total_ops
        new.max_per_unit_ops += other.max_per_unit_ops
        new.max_latency += other.max_latency
        # max_first_latency is only ever updated across loops ABOVE the loop
        # for which we calculated that first latency, so we should MAX
        new.max_first_latency = max_dict(
            self.max_first_latency, other.max_first_latency
        )  # FIRST LATENCY
        return new

    def combine_temporal(self, other: "ComputeStats"):
        self.total_ops += other.total_ops
        self.max_per_unit_ops += other.max_per_unit_ops
        self.max_latency += other.max_latency
        # max_first_latency is only ever updated across loops ABOVE the loop
        # for which we calculated that first latency, so we should MAX
        self.max_first_latency = max_dict(
            self.max_first_latency, other.max_first_latency
        )  # FIRST LATENCY

    def combine_spatial(self, other: "ComputeStats"):
        self.total_ops += other.total_ops
        self.max_per_unit_ops = MaxGeqZero(
            self.max_per_unit_ops, other.max_per_unit_ops
        )
        self.max_latency = MaxGeqZero(self.max_latency, other.max_latency)
        # max_first_latency is only ever updated across loops ABOVE the loop
        # for which we calculated that first latency, so we should MAX
        self.max_first_latency = max_dict(
            self.max_first_latency, other.max_first_latency
        )  # FIRST LATENCY


@dataclass
class SymbolicAnalysisOutput:
    compute_stats: dict[Compute, ComputeStats] = field(default_factory=dict)

    buffet_stats: dict[Buffet, BuffetStats] = field(default_factory=dict)

    network_stats: dict[Network, NetworkStats] = field(default_factory=dict)

    # Mapping [level, einsum] to the fanout
    fanout: dict[(Buffet, str), int] = field(default_factory=dict)

    # Mapping [einsum] to the number of temporal steps
    temporal_steps: dict[str, int] = field(default_factory=dict)

    symbols: list[sympy.Symbol] = field(default_factory=list)

    # tensor to the mapping for that particular tensor
    tensor2mapping: dict[TensorName, Mapping] = field(default_factory=dict)

    def get_buffet_for_tensor(self, tensor: TensorName) -> Buffet:
        for buffet in self.buffet_stats:
            if buffet.tensor == tensor:
                return buffet
        raise ValueError(f"Buffet for tensor {tensor} not found")

    def max(self, **kwargs: Any):
        for key, value in kwargs.items():
            assert key in [
                "compute_stats",
                "stats",
                "fanout",
                "temporal_steps",
            ]
            previous = getattr(self, key)
            for k, v in value.items():
                previous.setdefault(k, {})
                for k2, v2 in v.items():
                    if k2 in previous[k]:
                        previous[k][k2] = MaxGeqZero(previous[k][k2], v2)
                    else:
                        previous[k][k2] = v2

    def get_child_buffet_stats(self, buffet: Buffet) -> BuffetStats:
        seen = False
        for child_buffet, child_stats in reversed(self.buffet_stats.items()):
            if not seen:
                seen = child_buffet == buffet
                continue
            if child_buffet.tensor == buffet.tensor:
                return child_stats
        return None

    def sum_buffet_stats_per_level(self) -> dict[str, BuffetStats]:
        result: dict[str, BuffetStats] = {}
        for buffet, stats in self.buffet_stats.items():
            result.setdefault(buffet.level, BuffetStats.blank())
            result[buffet.level] += stats
        return result

    def add_buffet_stats_and_symbols(self, other: "SymbolicAnalysisOutput"):
        assert not (oset(self.buffet_stats) & oset(other.buffet_stats)), "BUG"
        self.buffet_stats.update(other.buffet_stats)
        # if self.temporal_steps != other.temporal_steps:
        #     print(f'Temporal steps are different.')
        #     print(f'\tmine:  {self.temporal_steps}')
        #     print(f'\tother: {other.temporal_steps}')
        # assert self.temporal_steps == other.temporal_steps, "BUG"
        self.temporal_steps.update(other.temporal_steps)
        self.symbols.extend([s for s in other.symbols if s not in self.symbols])

    def add_network_stats(self, other: "SymbolicAnalysisOutput"):
        assert not (oset(self.network_stats) & oset(other.network_stats)), "BUG"
        self.network_stats.update(other.network_stats)



import accelforge as af
import os

THIS_SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
VARIABLES_GLOBAL_PATH = os.path.join(THIS_SCRIPT_DIR, "_include.yaml")


def get_spec(
    arch_name: str,
    compare_with_arch_name: str | None = None,
    add_dummy_main_memory: bool = False,
    n_macros: int = 1,
) -> af.Spec:
    """
    Gets the spec for the given architecture. If `compare_with_arch_name` is given, the
    variables_iso will be grabbed from `compare_with_arch_name` in order to match
    attributes for fair comparison.

    Parameters
    ----------
    arch_name: str
        The name of the architecture to get the spec for.
    compare_with_arch_name: str | None
        The name of the architecture to compare with. If not given, variables will be
        taken from the given `arch_name`.
    n_macros: int
        The number of macros to use in the architecture.
    Returns
    -------
    spec: af.Spec
        The spec for the given architecture.
    """
    if compare_with_arch_name is None:
        compare_with_name = arch_name
    else:
        compare_with_name = compare_with_arch_name

    arch_name_base = arch_name
    arch_name = os.path.join(THIS_SCRIPT_DIR, f"{arch_name}.yaml")
    compare_with_name = os.path.join(THIS_SCRIPT_DIR, f"{compare_with_name}.yaml")
    variables = af.Variables.from_yaml(arch_name, top_key="variables")
    arch = af.Arch.from_yaml(arch_name, top_key="arch")
    workload = af.Workload.from_yaml(arch_name, top_key="workload")
    spec = af.Spec(arch=arch, variables=variables, workload=workload)

    spec.config.expression_custom_functions.append(
        os.path.join(THIS_SCRIPT_DIR, "_include_functions.py")
    )
    # Load architecture-specific helper functions if they exist
    arch_helpers = os.path.join(
        THIS_SCRIPT_DIR, f"{arch_name_base}_helper_functions.py"
    )
    if os.path.exists(arch_helpers):
        spec.config.expression_custom_functions.append(arch_helpers)
    spec.config.component_models.append(
        os.path.join(THIS_SCRIPT_DIR, "components/*.py")
    )
    if n_macros > 1:
        macro = af.arch.Container(
            name="MacroAuto",
            spatial=[{"name": "macro", "fanout": n_macros, "power_gateable": True}],
        )
        spec.arch.nodes.insert(0, macro)
    if add_dummy_main_memory:
        main_memory = af.arch.Memory(
            name="MainMemory",
            component_class="Dummy",
            size=float("inf"),
            tensors={"keep": "~weight"},
        )
        spec.arch.nodes.insert(0, main_memory)
    return spec



from collections.abc import Iterable
from pathlib import Path

from accelforge.util._picklecache import PickleCache

TEST_PmappingGroup_DIR = Path(__file__).parent / "tmp" / "pmapping_cache"


def make_pmapping_pickle_cache(config_names: Iterable[str]) -> PickleCache:
    """
    Return a `PickleCache` using path generated by `make_pmapping_pickle_path`.
    """
    return PickleCache(make_pmapping_pickle_path(config_names))


def make_pmapping_pickle_path(config_names: Iterable[str]) -> Path:
    """
    Creates a path to a file in `TEST_PmappingGroup_DIR` with a file name that is
    `config_names` joined using "---" as separators.
    """
    pickle_name = "---".join(sorted(config_names)) + ".pkl"
    pickle_path = TEST_PmappingGroup_DIR / pickle_name
    return pickle_path



"""
Tests the temporal reuse analysis in ISL.
"""

import unittest

import islpy as isl

from accelforge.model._looptree.reuse.isl.mapping_to_isl.types import (
    Occupancy,
    SpatialTag,
    TemporalTag,
)
from accelforge.model._looptree.reuse.isl.temporal import (
    TemporalReuse,
    analyze_temporal_reuse,
)


class TestTemporalReuseAnalysis(unittest.TestCase):
    """
    Tests the temporal reuse analysis in ISL.
    """

    def test_multiple_loop_reuse_sandbox(self):
        """
        Tests a very basic temporal reuse analysis case without needing the
        `occupancies_from_mapping` function.
        """
        # Loads in the CONV1D Config
        occ: Occupancy = Occupancy(
            [TemporalTag(), SpatialTag(0, None), TemporalTag()],
            isl.Map.read_from_str(
                isl.DEFAULT_CONTEXT,
                (
                    "{ generic_iteration[t1, x, t0] -> tensor[d] : "
                    "t0 <= d < t0+3 and 0 <= t1 < 2 and 0 <= x < 2 and 0 <= t0 < 2 }"
                ),
            ).coalesce(),
        )

        result: TemporalReuse = analyze_temporal_reuse(occ, True, True)
        soln: isl.Map = isl.Map.read_from_str(
            isl.DEFAULT_CONTEXT,
            "{ generic_iteration[t1, x, t0] -> tensor[d] : "
            "0 <= x < 2 and "
            "(((t1 = 0) and (t0 = 0) and (0 <= d < 3)) or "
            " ((t1 = 0) and (t0 = 1) and (d = 3)) or "
            " ((t1 = 1) and (t0 = 0) and (d = 0)) or "
            " ((t1 = 1) and (t0 = 1) and (d = 3))"
            ")}",
        ).coalesce()
        fill: isl.Map = result.fill.map_
        eff_occ: isl.Map = result.effective_occupancy.map_
        print(eff_occ)
        assert soln.is_equal(fill), f"Expected:\n{soln}\nReceived:\n{fill}"



from collections.abc import Collection, Generator, Sequence
from dataclasses import dataclass
from itertools import product
import logging
from typing import Any

import accelforge.frontend.arch as arch
from accelforge.frontend.mapping import MappingNode, Storage, Toll, TensorHolder
from accelforge.frontend.spec import Spec
from accelforge.frontend.workload import TensorName, SymbolTable
from accelforge.util._eval_expressions import MATH_FUNCS

from accelforge.util._frozenset import oset
from accelforge.mapper.FFM._make_pmappings.make_pmapping_templates.make_storages import (
    make_storage_choices_all_levels,
)
from accelforge.frontend.workload import EinsumName


def get_tensor_choices(
    einsum_name: EinsumName,
    nodes: list[arch.Memory],
    symbol_table: SymbolTable,
    spec: Spec,
    first_memory: arch.Memory,
    fusable_tensors: set[TensorName],
    fanouts: dict[str, int],
    prioritize_reuse_of_unfused_tensors: bool,
) -> Generator[tuple[list[TensorHolder], SymbolTable, arch.Compute], None, None]:
    nodes, compute = nodes[:-1], nodes[-1]
    nodes = list(filter(lambda n: isinstance(n, arch.TensorHolder), nodes))

    tensors = spec.workload.einsums[einsum_name].tensor_names
    is_copy_op = spec.workload.einsums[einsum_name].is_copy_operation
    persistent_tensors = oset(
        t.name
        for t in spec.workload.einsums[einsum_name].tensor_accesses
        if t.persistent
    )

    for choice, symbol_table in make_storage_choices_all_levels(
        nodes=nodes,
        symbol_table=symbol_table,
        is_copy_op=is_copy_op,
        persistent_tensors=persistent_tensors,
        seen_tensors=oset(),
        einsum_name=einsum_name,
        prioritize_reuse_of_unfused_tensors=prioritize_reuse_of_unfused_tensors,
    ):
        x = [y for z in choice.values() for y in z]
        logging.info(
            f"\t\tUnordered storage choice: {", ".join(n.compact_str() for n in x)}"
        )
        all_tensor_holders = [v2 for v in choice.values() for v2 in v]
        storage_holders = [th for th in all_tensor_holders if isinstance(th, Storage)]
        toll_holders = [th for th in all_tensor_holders if isinstance(th, Toll)]

        # Start out the mapping with the outermost memory name
        base_mapping = []

        # make_storage_chocies_all_levels will parse the "tensors" field for each node,
        # and we'll want those parsed values for later steps. This code grabs it from
        # the nodes in the mapping. We only check the first storage node for each memory
        # level since they're all the same.
        parsed_nodes_in_mapping = []
        for k, v in choice.items():
            for s in v[:1]:
                parsed_nodes_in_mapping.append(s.component_object)

        # Get the dataflow constraints for the mapping
        required_order = get_tensor_order_constraint(
            parsed_nodes_in_mapping, symbol_table, tensors
        )

        symbol_table["arch_attributes"] = {}
        cur_compute = compute._eval_expressions(
            symbol_table,
            location=f"arch.{compute.name}",
            musteval_tryeval_to=True,
            must_copy=False,
        )[0]
        assert isinstance(cur_compute.enabled, bool)
        if not cur_compute.enabled:
            continue

        for mapping in recursive_order_tensor_choices(
            einsum_name,
            tensors,
            base_mapping,
            nodes,
            storage_holders,
            required_order,
            spec,
            is_copy_op,
            first_memory,
            fusable_tensors,
            fanouts,
        ):
            mapping = insert_tolls(mapping, toll_holders, nodes, fanouts)
            yield mapping, symbol_table, cur_compute


def get_tensor_order_constraint(nodes, symbol_table, tensors):
    required_order: dict[str, list[Order]] = {}
    seen_tensors = oset()
    for node in nodes:
        if isinstance(node, arch.Container):
            continue
        node_tensors: arch.Tensors = node.tensors._eval_expressions(
            symbol_table=symbol_table,
            musteval_tryeval_to=True,
            must_copy=False,
            location=f"arch.{node.name}.tensors",
        )[0]
        tensor_order_options = list(node_tensors.tensor_order_options)
        for order_constraint in tensor_order_options:
            order = Order()
            for together_tensors in order_constraint:
                in_mapping_together_tensors = [
                    tensor for tensor in together_tensors if tensor in tensors
                ]
                if len(in_mapping_together_tensors) == 1:
                    only_tensor = in_mapping_together_tensors[0]
                    order.add_tensor(only_tensor)
                elif len(in_mapping_together_tensors) > 1:
                    order.add_together_tensors(in_mapping_together_tensors)
            if order.order:
                required_order.setdefault(node.name, []).append(order)
        seen_tensors.update(node_tensors.keep)
    return required_order


def insert_tolls(
    mapping: list[TensorHolder],
    toll_holders: list[Toll],
    arch_nodes: list[arch.TensorHolder],
    fanouts: dict[str, int],
) -> list[MappingNode]:
    if not toll_holders:
        return mapping

    mapping = list(mapping)

    arch_order = {n.name: i for i, n in enumerate(arch_nodes)}

    # Sort tolls by arch order, then by tensor name
    toll_holders = sorted(
        toll_holders,
        key=lambda t: (arch_order[t.component], sorted(t.tensors)),
    )

    for toll in toll_holders:
        toll_arch_idx = arch_order[toll.component]
        toll_fanout = fanouts.get(toll.component, 0)

        # Must go below the storage node above them
        toll_tensors = oset(toll.tensors)
        min_pos = 0
        for i, m in enumerate(mapping):
            if (
                oset(m.tensors) & toll_tensors
                and arch_order[m.component] < toll_arch_idx
            ):
                min_pos = max(min_pos, i + 1)

        # Rule 2: Must go above the storage node below them
        max_pos = len(mapping)
        for i, m in enumerate(mapping):
            if (
                oset(m.tensors) & toll_tensors
                and arch_order[m.component] > toll_arch_idx
            ):
                max_pos = min(max_pos, i)
                break

        assert min_pos <= max_pos

        # If possible, go below fanout above them
        for i in range(min_pos, min(max_pos, len(mapping))):
            if fanouts.get(mapping[i].component, 0) < toll_fanout:
                min_pos = i + 1

        # Go below any already-inserted entry from an arch level above this toll, and
        # maintain alphabetical order for tolls at the same arch level
        for i in range(min_pos, min(max_pos, len(mapping))):
            m = mapping[i]
            m_arch_idx = arch_order.get(m.component, -1)
            if m_arch_idx < toll_arch_idx:
                min_pos = i + 1
            elif m_arch_idx == toll_arch_idx and sorted(m.tensors) < sorted(
                toll.tensors
            ):
                min_pos = i + 1

        # Rule 4: Go as high as possible
        mapping.insert(min_pos, toll)

    return mapping


def recursive_order_tensor_choices(
    einsum_name: EinsumName,
    tensors: set[TensorName],
    mapping: Sequence[MappingNode],
    nodes: list[arch.Memory],
    remaining_choices: list,
    required_order: list[list[TensorHolder]],
    spec: Spec,
    is_copy_op: bool,
    first_memory: arch.Memory,
    fusable_tensors: set[TensorName],
    fanouts: dict[str, int],
) -> Generator[list[MappingNode], None, None]:
    def check_has_tensors(mapping: list[MappingNode]):
        tensor_holders = [node for node in mapping if isinstance(node, TensorHolder)]
        tensors_in_mapping = oset(
            tensor
            for tensor_holder in tensor_holders
            for tensor in tensor_holder.tensors
        )
        if tensors_in_mapping != tensors:
            raise ValueError(
                f"Einsum {einsum_name} has a pmapping template that is missing tensors. Ensure "
                f"that there is a storage node storing each tensor in the Einsum. Missing "
                f"tensors: {tensors - tensors_in_mapping}. Pmapping template:\n\t"
                + "\n\t".join(m.compact_str() for m in mapping)
            )

    mapping = list(mapping)
    if not remaining_choices:
        check_has_tensors(mapping)
        yield mapping
        return

    # If it's a copy op and we have the backing storage for every tensor, return
    # immediately
    if is_copy_op:
        tensor_holders = [node for node in mapping if isinstance(node, TensorHolder)]
        if oset().union(*[t._backing for t in tensor_holders]) == tensors:
            check_has_tensors(mapping)
            yield mapping
            return

    for choice in sorted(remaining_choices, key=lambda x: x.compact_str()):
        mapping.append(choice)
        new_remaining = [c for c in remaining_choices if c != choice]
        valid, reason = valid_tensor_holder_order(
            mapping,
            [n.name for n in nodes],
            required_order,
            spec,
            first_memory,
            fusable_tensors,
            fanouts,
            new_remaining,
        )
        if valid:
            yield from recursive_order_tensor_choices(
                einsum_name,
                tensors,
                mapping,
                nodes,
                new_remaining,
                required_order,
                spec,
                is_copy_op,
                first_memory,
                fusable_tensors,
                fanouts,
            )
        else:
            logging.info(
                "\t\t"
                + " " * len(mapping)
                + f"Invalid tensor holder order: {", ".join(n.compact_str() for n in mapping)}: {reason}"
            )
        mapping.pop()


def valid_tensor_holder_order(
    mapping: Sequence[TensorHolder],
    node_names: list[str],
    required_orders: dict[str, list["Order"]],
    spec: Spec,
    first_memory: arch.Memory,
    fusable_tensors: set[TensorName],
    fanouts: dict[str, int],
    remaining_choices: list[TensorHolder],
):
    memory_to_satisfied_constraints: dict[str, set] = {}

    for i, m0 in enumerate(mapping):
        # There are some checks for back-to-back nodes, but ones in remaining_choices
        # aren't yet ordered so we can't decide if anything is directly back-to-back.
        # However we can check longer-distance relations between nodes in our mapping
        # and remaining choices because we know that they will be placed somewhere
        # eventually. So we'll add a None between to act as a separator, then we'll
        # never get straight back-to-back nodes between the mapping and remaining
        # choices.
        following = mapping[i + 1 :] + [None] + remaining_choices
        for j, m1 in enumerate(following):
            if following[j] is None:
                continue

            j += i + 1

            s1, s2 = m0.component, m1.component
            s1_idx, s2_idx = node_names.index(s1), node_names.index(s2)
            s1_persistent, s2_persistent = m0.persistent, m1.persistent
            either_persistent = s1_persistent or s2_persistent

            assert len(m0.tensors) == 1
            assert len(m1.tensors) == 1

            # If they're persistent they're forced to be at the top.
            force_order = (
                spec.mapper.force_memory_hierarchy_order and not either_persistent
            )
            force_order &= m0.component_object.tensors.force_memory_hierarchy_order
            force_order &= m1.component_object.tensors.force_memory_hierarchy_order

            # Ctrl-F for CONTIGUOUS_ITERATION_SPACE_DISCUSSION: The following line does
            # not let backing storage be above in the mapping anything that is below it
            # in the memory hierarchy. THIS IS NOT FUNDAMENTAL. If we remove this
            # constraint, then the fused loops may be different across different backing
            # storages, so we would need to update make_pmappings_from_templates.py to
            # make compatibility from the mapping for each tensor.
            force_order |= bool(m0._backing & fusable_tensors)

            if force_order and i < j and s2_idx < s1_idx:
                return (
                    False,
                    f"Memory {s1} is below memory {s2}, violating memory hierarchy order.",
                )

            s1_outermost = s1_persistent
            s2_outermost = s2_persistent
            if not spec.mapper._can_lower_outermost_memory:
                s1_outermost |= s1 == first_memory.name
                s2_outermost |= s2 == first_memory.name

            # Persistent tensors must be at the top of the hierarchy
            if s2_outermost and not s1_outermost and i < j:
                return (
                    False,
                    f"Outermost {m0.compact_str()}, persistent {s1_persistent} is below non-outermost {m1.compact_str()}, persistent {s2_persistent}.",
                )

            # If they're both the first memory and we can't lower the outermost memory,
            # then relative order doesn't matter because we can't have any loops between
            # them.
            if (
                s1 == first_memory.name
                and s2 == first_memory.name
                and not spec.mapper._can_lower_outermost_memory
            ):
                if sorted(m0.tensors) < sorted(m1.tensors):
                    return (
                        False,
                        f"Force alphabetical order for storage in outermost memory. {m0.compact_str()} is before {m1.compact_str()}.",
                    )
                # Ignore the following constraints
                continue

            if s1 == s2 and s1 in required_orders and i != j:
                if s1 not in memory_to_satisfied_constraints:
                    memory_to_satisfied_constraints[s1] = oset(
                        i for i in range(len(required_orders[s1]))
                    )

                good = True
                for order_idx, order_choice in enumerate(required_orders[s1]):
                    if order_idx not in memory_to_satisfied_constraints[s1]:
                        continue

                    good = True
                    for t1, t2 in product(m0.tensors, m1.tensors):
                        idx_of_i_in_order = order_choice.index(t1)
                        idx_of_j_in_order = order_choice.index(t2)

                        if idx_of_i_in_order is None or idx_of_j_in_order is None:
                            continue

                        if idx_of_i_in_order > idx_of_j_in_order:
                            good = False
                            reason = f"Tensor {t1} is before tensor {t2} in the order {order_choice}"
                            break
                    if not good:
                        memory_to_satisfied_constraints[s1].remove(order_idx)

                if len(memory_to_satisfied_constraints[s1]) == 0:
                    return False, reason

            if not (oset(m0.tensors) & oset(m1.tensors)):
                continue

            if i < j and s2_idx < s1_idx:
                return False, f"{m0.compact_str()} is below {m1.compact_str()}"

            # If a tensor is stored in two levels back-to-back, then we should have
            # bypassed the outer TensorHolder if possible.
            either_backing = m0._backing & m1._backing
            same_fanout = fanouts[s1] == fanouts[s2]
            if (
                "redundant_dataplacements"
                not in spec.mapper._count_option_for_mapsapce_size_evaluation
            ):
                if same_fanout and (i == j or i == j - 1):
                    if s1_idx < s2_idx and not (
                        (oset(m0._must_keep_tensors) & oset(m1.tensors))
                        or either_backing
                    ):
                        shared = oset(m0._must_keep_tensors) & oset(m1.tensors)
                        return (
                            False,
                            f"{shared} stored in back-to-back storage nodes, and could have bypassed the outer one.",
                        )
                    if s2_idx < s1_idx and not (
                        (oset(m1._must_keep_tensors) & oset(m0.tensors))
                        or either_backing
                    ):
                        shared = oset(m1._must_keep_tensors) & oset(m0.tensors)
                        return (
                            False,
                            f"{shared} is stored in back-to-back storage nodes, and could have bypassed the outer one.",
                        )

    for i, m0 in enumerate(mapping):
        for j, m1 in enumerate(mapping[i:]):
            s1, s2 = m0.component, m1.component
            if s1 != s2 or s1 not in memory_to_satisfied_constraints or i == j:
                continue

            satisfied_orders = memory_to_satisfied_constraints[s1]
            assert len(satisfied_orders) > 0

            for order_idx in satisfied_orders:
                order = required_orders[s1][order_idx]
                for tensor_i in m0.tensors:
                    for tensor_j in m1.tensors:
                        if order.index(tensor_i) != order.index(tensor_j):
                            continue
                break

    return True, ""


@dataclass(frozen=True)
class Alone:
    tensor: Any


@dataclass(frozen=True)
class Together:
    tensors: Collection[Any]


class Order:
    """An ordering of tensors."""

    def __init__(self):
        self.order = []

    def __repr__(self):
        return f"Order({self.order})"

    def add_tensor(self, tensor):
        self.order.append(Alone(tensor))

    def add_together_tensors(self, together_tensors):
        self.order.append(Together(together_tensors))

    def index(self, tensor):
        for i, order_term in enumerate(self.order):
            if (isinstance(order_term, Alone) and order_term.tensor == tensor) or (
                isinstance(order_term, Together) and tensor in order_term.tensors
            ):
                return i
        return None



import accelforge as af


def get_array_fanout_reuse_input(spec: af.Spec) -> int:
    """Get total fanout of array spatial dims that reuse input (= columns)."""
    n = 1
    for leaf in spec.arch.get_nodes_of_type(af.arch.Leaf):
        for sp in leaf.spatial:
            if sp.name.endswith("ARRAY_COLUMNS") or sp.name.endswith("ARRAY_ROWS"):
                if str(sp.may_reuse) == "input" or str(sp.reuse) == "input":
                    n *= sp.fanout
    return n


def get_array_fanout_reuse_output(spec: af.Spec) -> int:
    """Get total fanout of array spatial dims that reuse output (= rows)."""
    n = 1
    for leaf in spec.arch.get_nodes_of_type(af.arch.Leaf):
        for sp in leaf.spatial:
            if sp.name.endswith("ARRAY_COLUMNS") or sp.name.endswith("ARRAY_ROWS"):
                if str(sp.may_reuse) == "output" or str(sp.reuse) == "output":
                    n *= sp.fanout
    return n


def get_array_fanout_total(spec: af.Spec) -> int:
    return get_array_fanout_reuse_input(spec) * get_array_fanout_reuse_output(spec)


# Sign magnitude
# 1. Scale X to [-1, 1]: x = NORM_1_TO_NEG1(x, INPUTS_VALUE_DISTRIBUTION)
# 2. Convert to signed: x = abs(x) * (2 ** (INPUT_BITS - 1) - 1)
# 2. x = round(x * (2 ** INPUT_BITS - 1))

from math import log2
from typing import List, NamedTuple, Union


class ProbableBits(NamedTuple):
    bits: list
    probability: float


# ==============================================================================
# Encoding functions
# ==============================================================================


def magnitude_encode_hist(weights) -> List[ProbableBits]:
    """
    A signed value is encoded as a positive or negative magnitude of that value.
    Signed hardware is requireed.
    """
    nbits = get_n_bits(weights)
    encoded = []
    halfwidth = len(weights) / 2
    for i, w in enumerate(weights):
        normed = norm(i, len(weights), -halfwidth + 0.5, halfwidth + 0.5)
        encoded.append(ProbableBits(to_bits_unsigned(abs(normed), nbits)[1:], w))
    return norm_encoded_hist(encoded)


def two_part_magnitude_encode_hist(weights):
    """
    Two (devices, timesteps, components, etc.) encode each signed value. If the
    value is positive, the first device encodes the magnitude of the value. If the
    value is negative, the second device encodes the magnitude of the value. The
    other device encodes 0.
    """
    m = magnitude_encode_hist(weights)
    m2 = []
    for e in m:
        m2.append(ProbableBits(e.bits, e.probability / 2))
        m2.append(ProbableBits([0] * len(e.bits), e.probability / 2))
    return m2


def offset_encode_hist(weights):
    """
    A signed value is encoded as the the value minus the negative minimum value.
    This maps a range of [-min, max] to [0, max - min]. The bias must be added
    back after computation.
    """
    nbits = get_n_bits(weights)
    encoded = []
    for i, w in enumerate(weights):
        normed = norm(i, len(weights), 0, len(weights))
        encoded.append(ProbableBits(to_bits_unsigned(normed, nbits), w))
    return norm_encoded_hist(encoded)


def offset_encode_if_signed_hist(weights):
    """
    Offset encode a value only if it is signed. Otherwise, don't apply any bias and just
    use the positive values.
    """
    if is_hist_signed(weights):
        return offset_encode_hist(weights)
    return magnitude_encode_hist(weights)


def two_part_magnitude_encode_if_signed_hist(weights):
    """
    Two part magnitude encode a value only if it is signed. Otherwise, use only posiive
    values.
    """
    if is_hist_signed(weights):
        return two_part_magnitude_encode_hist(weights)
    return magnitude_encode_hist(weights)


def xnor_encode_hist(weights):
    """
    XNOR encoding based on Jia JSSCC 2020.
    """
    nbits = get_n_bits(weights)
    encoded = []
    halfwidth = len(weights) / 2
    for i, w in enumerate(weights):
        normed = norm(i, len(weights), -halfwidth + 0.5, halfwidth + 0.5)
        bits = []
        for j in list(range(nbits - 1, -1, -1)) + [-1, -1]:
            bits.append(int(normed > 0))
            normed -= 2**j * (2 * bits[-1] - 1)
        assert normed == 0, f"normed={normed} is not 0"
        encoded.append(ProbableBits(bits, w))
    return norm_encoded_hist(encoded)


def zero_gated_xnor_encode_hist(weights):
    """
    XNOR encoding with zero gating based on Jia JSSCC 2020.
    """
    encoded = xnor_encode_hist(weights)
    zero_idx = len(encoded) // 2
    encoded[zero_idx] = ProbableBits(
        [0] * len(encoded[zero_idx].bits), encoded[zero_idx].probability
    )
    return encoded


# ==============================================================================
# Helper functions
# ==============================================================================


def assert_hist_pow2_minus1(hist):
    x = 1
    while x <= len(hist):
        x *= 2
    assert x - 1 == len(
        hist
    ), f"Histogram length {len(hist)} is not a power of 2 minus 1."


def norm_encoded_hist(encoded_hist: List[ProbableBits]):
    sum_probs = sum([e.probability for e in encoded_hist])
    return [ProbableBits(e.bits, e.probability / sum_probs) for e in encoded_hist]


def get_n_bits(hist):
    n_bits = 0
    while 2**n_bits < len(hist) + 1:
        n_bits += 1
    assert (
        2**n_bits == len(hist) + 1
    ), f"Number of histogram bins + 1 must be a power of 2, got {len(hist)}."
    return n_bits


def is_hist_signed(hist):
    return sum(hist[: len(hist) // 2]) != 0


def hist_to_magnitude(hist):
    assert_hist_pow2_minus1(hist)
    new_hist = [0] * (len(hist) // 2)
    hist_center = len(hist) // 2
    for i in range(len(new_hist)):
        new_hist[i] = hist[hist_center + i] + hist[hist_center - i]
    assert_hist_pow2_minus1(new_hist)
    return new_hist


def to_bits_unsigned(x, nbits):
    x = round(x)
    assert 0 <= x < 2**nbits, f"x={x} is not in range [0, 2^{nbits})"
    return [int(i) for i in bin(x)[2 : nbits + 2].zfill(nbits)]


def norm(x, nbins, rmin, rmax):
    return x / nbins * (rmax - rmin) + rmin


def encoded_hist_to_avg_slice(
    encoded_hist: List[ProbableBits],
    total_bits: int,
    bits_per_slice: Union[list, int],
    partial_slices_use_full_range: bool = False,
    return_per_slice: bool = False,
):
    if isinstance(bits_per_slice, int):
        bits_per_slice = [bits_per_slice] * (total_bits // bits_per_slice)
        if sum(bits_per_slice) != total_bits:
            bits_per_slice.append(total_bits - sum(bits_per_slice))

    assert total_bits == sum(bits_per_slice), (
        f"Sum of bits per slice {sum(bits_per_slice)} != total_bits " f"{total_bits}"
    )

    bit2slice = []
    max_val = max(2 ** max(bits_per_slice) - 1, 1)
    for i, b in enumerate(bits_per_slice):
        m = max(2**b - 1, 1) if partial_slices_use_full_range else max_val
        bit2slice += [(i, max((2 ** (b - j - 1)), 1) / m) for j in range(b)]

    avg_slice_values = [0] * len(bits_per_slice)
    for e in encoded_hist:
        for i in range(total_bits):
            slice_idx, scale = bit2slice[i]
            if i >= len(e.bits):
                bit_value = sum(e.bits) / len(e.bits)
            else:
                bit_value = e.bits[i]
            avg_slice_values[slice_idx] += bit_value * e.probability * scale

    if return_per_slice:
        return avg_slice_values

    return sum(avg_slice_values) / len(avg_slice_values)


if __name__ == "__main__":
    input_dist = [16 - abs(16 - i) for i in range(31)]
    print(f"input_dist: {input_dist}")
    for e in xnor_encode_hist(input_dist):
        print(e)



from accelforge.frontend.mapper.metrics import Metrics
from accelforge.util._basetypes import EvalableModel


class Model(EvalableModel):
    """Configuration for the model."""

    metrics: Metrics = Metrics.all_metrics()
    """
    Metrics to evaluate.

    If using spec to call mapper, leave this configuration as is. The mapper
    will make necessary configurations.
    """



from numbers import Number
from typing import Optional, List
from hwcomponents.scaling import tech_node_area
from util.bit_functions import rescale_sum_to_1
from hwcomponents import ComponentModel, action


class Capacitor(ComponentModel):
    """
    A capacitor.

    Parameters
    ----------
    capacitance: float
        The capacitance of this capacitor in Farads.
    tech_node: int
        The tech_node node in meters.
    voltage: float
        The supply voltage in volts.
    cap_per_m2: float
        The capacitance per square meter in Farads per square meter.
    border_area: float
        The border area around the capacitor in square meters.
    voltage_raise_threshold: float
        Latency is calculated as the time it takes to raise voltage to this proportion
        of the target voltage.
    supply_resistance: float
        The supply resistance in ohms. If 0, then voltage is assumed to converge
        instantly.

    Attributes
    ----------
    capacitance: float
        The capacitance of this capacitor in Farads.
    tech_node: int
        The tech_node node in meters.
    voltage: float
        The supply voltage in volts.
    cap_per_m2: float
        The capacitance per square meter in Farads per square meter.
    border_area: float
        The border area around the capacitor in square meters.
    voltage_raise_threshold: float
        Latency is calculated as the time it takes to raise voltage to this proportion
        of the target voltage.
    supply_resistance: float
        The supply resistance in ohms. If 0, then voltage is assumed to converge
        instantly.
    """

    priority = 0.5
    """
    Priority determines which model is used when multiple models are available for a
    given component. Higher priority models are used first. Must be a number between 0
    and 1.
    """

    def __init__(
        self,
        capacitance: Number,
        tech_node: float,
        voltage: Number = 0.7,
        cap_per_m2: Optional[Number] = "1e-3 scaled by tech node",
        border_area: Optional[Number] = "1e-12 scaled by tech node",
    ):
        self.capacitance = capacitance
        self.voltage = voltage

        if cap_per_m2 == "1e-3 scaled by tech node":
            cap_per_m2 = 2.3e-3 * tech_node_area(tech_node, 22e-9)
        if border_area == "1e-12 scaled by tech node":
            border_area = 1e-12 * tech_node_area(tech_node, 22e-9)

        self.cap_per_m2 = cap_per_m2
        self.border_area = border_area

        super().__init__(
            area=self.capacitance / self.cap_per_m2 + self.border_area, leak_power=0
        )

    @action
    def raise_voltage_to(
        self,
        target_voltage: float,
        supply_voltage: float = None,
    ) -> float:
        """
        Raise the voltage to the target voltage using the supply voltage as a supply.

        Parameters
        ----------
        target_voltage: float
            The target voltage to raise the voltage to.
        supply_voltage: float
            The supply voltage to use as a supply. If None, then the supply voltage is
            assumed to be the voltage set in the attributes of this capacitor.

        Returns
        -------
        energy, latency: tuple
            The energy required to raise the voltage to the target voltage. Latency is
            0.
        """
        if supply_voltage is None:
            supply_voltage = self.voltage
        assert target_voltage <= supply_voltage, (
            f"Can not raise voltage to {target_voltage} when supply voltage "
            f"is {supply_voltage}."
        )
        return self.capacitance * target_voltage * supply_voltage, 0

    @action
    def switch(
        self,
        value_probabilities: List[Number],
        zero_between_values: bool = True,
        supply_voltage: float = None,
    ) -> float:
        """
        Calculates the expected energy to switch voltage to the values in
        value_probabilities.

        Parameters
        ----------
        value_probabilities: List[Number]
            The probabilities of the values to switch to. This is a histogram, assumed
            to be spaced between 0 and supply_voltage, inclusive.
        zero_between_values: bool
            Whether to zero the voltage between values.
        supply_voltage: float
            The supply voltage to use as a supply. If None, then the supply voltage is
            assumed to be the voltage set in the attributes of this capacitor.

        Returns
        -------
        energy, latency: tuple
            The energy required to switch the voltage to the values in
            value_probabilities. Latency is 0.
        """
        supply_voltage = self.voltage if supply_voltage is None else supply_voltage
        expected_energy = 0
        value_probabilities = rescale_sum_to_1(value_probabilities)
        for v0, p0 in enumerate(value_probabilities):
            for v1, p1 in enumerate(value_probabilities):
                v0 = 0 if zero_between_values else v0
                if v1 < v0:
                    continue
                e0 = self.raise_voltage_to(
                    v0 / (len(value_probabilities) - 1) * self.voltage, supply_voltage
                )[0]
                e1 = self.raise_voltage_to(
                    v1 / (len(value_probabilities) - 1) * self.voltage, supply_voltage
                )[0]
                expected_energy += (e1 - e0) * p0 * p1
        return expected_energy, 0


class Wire(Capacitor):
    """
    A wire.

    Parameters
    ----------
    length: Number
        The length of the wire in meters.
    capacitance_per_m: Number
        The capacitance per meter in Farads per meter.
    voltage: Number
        The supply voltage of the wire in volts.

    Attributes
    ----------
    length: Number
        The length of the wire in meters.
    capacitance_per_m: Number
        The capacitance per meter in Farads per meter.
    voltage: Number
        The supply voltage of the wire in volts.
    """

    def __init__(
        self,
        length: Number,
        capacitance_per_m: Number = 2e-10,
        voltage: Number = 0.7,
        **kwargs,
    ):
        super().__init__(
            capacitance=length * capacitance_per_m,
            voltage=voltage,
        )
        self.length = length
        self.capacitance_per_m = capacitance_per_m
        self.voltage = voltage
        self.area_scale = 0
